import pandas as pd
import pytest

from ctdcal import process_bottle

# def test_add_btlnbr_cols():
#     df = {
#         "btl_fire_num": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
#         "other_id":     [1, 2, 3, 4, 5, 6, 10, 11, 12, 13, 14, 15]    #   Like the LADCP is mounted on it
#     }
#     df = pd.DataFrame(df)
#     btl_num_col = "btl_fire_num"

#     #   Call the function
#     result_df = process_bottle.add_btlnbr_cols(df, btl_num_col)

#     #   Check if the columns are added right
#     assert "BTLNBR" in result_df.columns
#     assert "BTLNBR_FLAG_W" in result_df.columns
#     assert all(result_df["BTLNBR_FLAG_W"] == 2)
#     assert result_df["BTLNBR"].equals(df[btl_num_col])

#     #   Cross compare
#     assert result_df["BTLNBR"].iloc[-1] != result_df["other_id"].iloc[-1]


def test_load_hy_file(tmp_path):
    # Create a sample file
    sample_data = """BOTTLE, SomecruiseODF
EXPOCODE,SECT_ID,STNNBR,CASTNO,SAMPNO,BTLNBR,BTLNBR_FLAG_W,DATE,TIME,LATITUDE,LONGITUDE,DEPTH,CTDPRS,CTDTMP,CTDSAL,CTDSAL_FLAG_W,SALNTY,SALNTY_FLAG_W,CTDOXY,CTDOXY_FLAG_W,OXYGEN,OXYGEN_FLAG_W,REFTMP,REFTMP_FLAG_W
,,,,,,,,,,,METERS,DBAR,ITS-90,PSS-78,,PSS-78,,UMOL/KG,,UMOL/KG,,ITS-90,
Somecruise,GOSHIP,1,1,36,36,2,19990722,1345,-32.69,114.97,114,3.5,20.30,35.4,2,-999.0,9,283.8,2,-999.0,9,-999.0,9
Somecruise,GOSHIP,1,1,35,35,2,19990722,1345,-32.69,114.97,114,4.4,20.30,35.4,2,-999.0,9,311.7,2,-999.0,9,20.3,2
Somecruise,GOSHIP,1,1,34,34,2,19990722,1345,-32.69,114.97,114,2.8,20.29,35.4,2,-999.0,9,274.8,2,-999.0,9,-999.0,9

END_DATA
    """
    fname = tmp_path / "sample_hyfile.csv"
    with open(fname, "w") as f:
        f.write(sample_data)

    df = process_bottle.load_hy_file(fname)

    #   Make sure everything is inside that we're expecting
    assert not df.empty
    assert "EXPOCODE" in df.columns
    assert "SECT_ID" in df.columns
    assert "STNNBR" in df.columns
    assert "CASTNO" in df.columns
    assert "DEPTH" in df.columns

    #   Check if the final row with EXPOCODE "END_DATA" is dropped
    assert "END_DATA" not in df["EXPOCODE"].values
    #   Value check
    assert df["BTLNBR"].iloc[-1] == 34


def test_load_btl_data(tmp_path):
    fname = tmp_path / "90909_btl_mean.pkl"
    df = pd.DataFrame(
        {
            "CTDTMP1": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
            "CTDTMP2": [0.5, 1.5, 2.5, 3.5, 4.5, 5.5],
            "CTDPRS": [5000, 4000, 3000, 2000, 1000, 0],
            "btl_fire_num": [1, 2, 3, 4, 5, 6],
        }
    )
    df.to_pickle(fname)

    # test read all columns
    btl_data = process_bottle._load_btl_data(fname)
    assert df.equals(btl_data[df.columns])
    assert btl_data.shape == (6, 5)
    assert all(btl_data["SSSCC"] == "90909")

    # test read select columns
    cols = ["CTDTMP1", "CTDPRS", "btl_fire_num"]
    btl_data_select = process_bottle._load_btl_data(fname, cols=cols)
    assert "CTDTMP2" not in btl_data_select.columns
    assert all(btl_data["SSSCC"] == "90909")


def test_load_reft_data(tmp_path):
    fname = tmp_path / "90909_reft.csv"
    df = pd.DataFrame(
        {
            "btl_fire_num": [1, 2, 3, 4, 5, 6],
            "T90": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
            "REFTMP_FLAG_W": 2,
        }
    )
    df.to_csv(fname)

    # test read
    reft_data = process_bottle._load_reft_data(fname)
    assert df.equals(reft_data[df.columns])
    assert all(reft_data["SSSCC_TEMP"] == "90909")
    assert reft_data["REFTMP"].equals(reft_data["T90"])


def test_load_salt_data(tmp_path):
    fname = tmp_path / "90909_salts.csv"
    df = pd.DataFrame(
        {
            "SAMPNO": [1, 2, 3, 4, 5, 6],
            "SALNTY": [30.0, 31.0, 32.0, 33.0, 34.0, 35.0],
            "BathTEMP": 24,
            "CRavg": [0.9, 0.91, 0.92, 0.93, 0.94, 0.95],
        }
    )
    df.to_csv(fname)

    # test read
    salt_data = process_bottle._load_salt_data(fname)
    assert all(salt_data["SSSCC_SALT"] == "90909")


@pytest.fixture
def df1():
    return pd.DataFrame(
        {
            "STNNBR": [101, 102, 100],
            "CASTNO": [1, 1, 1],
            "SAMPNO": [10, 11, 16],
            "TEMP": [20.5, 20.3, 19.6],
            "COND": [35.0, 34.9, 36.1],
            "OXY": [140.1, 128.0, 107.7],
        }
    )


@pytest.fixture
def df2():
    return pd.DataFrame(
        {
            "STNNBR": [103, 104],
            "CASTNO": [1, 1],
            "SAMPNO": [12, 13],
            "TEMP": [20.1, 19.8],
            "COND": [35.1, 35.2],
            "OXY": [108.1, 107.9],
            "XMISS": [90.5, 89.8],
        }
    )


def test_merge_hy1(df1, df2):
    merged_df = process_bottle.merge_hy1(df1, df2)

    #   Test for matching columns, dimensions, and contents
    assert list(merged_df.columns) == [
        "STNNBR",
        "CASTNO",
        "SAMPNO",
        "TEMP",
        "COND",
        "OXY",
        "XMISS",
    ]
    assert len(merged_df) == len(df1) + len(df2)
    assert not pd.isna(merged_df["XMISS"].iloc[-1])

    #   Test the sorting
    assert merged_df["STNNBR"].is_monotonic_increasing
    assert merged_df["CASTNO"].is_monotonic_increasing
    #   Sample number can be anything in this case

    #   Test handling NaNs
    df1.loc[0, "STNNBR"] = None
    merged_df = process_bottle.merge_hy1(df1, df2)
    assert merged_df["STNNBR"].isna().sum() == 1
    assert merged_df["SAMPNO"].notna().all()  #   No NaNs introduced



import json
from pathlib import Path

import pandas as pd
import pytest

from ctdcal.fitting.common import BottleFlags, df_node_to_BottleFlags, get_node, save_node, NodeNotFoundError


class TestBottleFlags:
    @pytest.fixture
    def sample_data(self, tmp_path):
        fname = tmp_path / 'data.json'
        with open(fname, 'w') as f:
            json.dump({'spam': {'egg': []}}, f)
        return fname

    @pytest.fixture
    def sample_fname(self, tmp_path):
        fname = tmp_path / 'save.json'
        return fname

    def test_add_node(self):
        flags = BottleFlags()
        flags.add_node('spam', ['egg', 'baked beans'])
        assert 'spam' in flags
        assert 'baked beans' in flags.spam
        assert len(flags.spam.egg) == 0

    def test_update_node(self, sample_data):
        with open(sample_data, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        assert 'lovely spam' not in flags.spam.egg
        # Test existing key
        flags.spam.update_node(egg='lovely spam')
        assert 'lovely spam' in flags.spam.egg
        # Test non-existing key
        with pytest.raises(KeyError):
            flags.spam.update_node(cheese='cheddar')

    def test_save(self, sample_data, sample_fname):
        with open(sample_data, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        assert not sample_fname.is_file()
        flags.save(sample_fname)
        assert sample_fname.is_file()


class TestBottleFlagWrangling:
    @pytest.fixture
    def sample_data(self, tmp_path):
        fname = tmp_path / 'data.json'
        with open(fname, 'w') as f:
            json.dump({'spam': {'egg': ['a', 'b'], 'sausage': ['C', 'D']}}, f)
        return fname

    @pytest.fixture
    def sample_df(self):
        spam_dict = {'spam': {'egg': ['a', 'b'], 'sausage': ['C', 'D']}}
        spam_df = pd.DataFrame.from_dict(spam_dict['spam'])
        return spam_df

    def test_df_node_to_BottleFlag(self, sample_df):
        spam = df_node_to_BottleFlags(sample_df)
        assert type(spam.sausage) is list
        assert spam.sausage[0] == 'C'

    def test_get_node(self, sample_data):
        # Test good node
        node = get_node(sample_data, 'spam')
        assert node.sausage[0] == 'C'
        # Test bad node
        with pytest.raises(NodeNotFoundError):
            get_node(sample_data, 'cheese')

    def test_save_node(self, sample_data, tmp_path):
        fname = Path(tmp_path, 'sample_copy.json')
        with open(sample_data, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        flags.save(fname)
        spam = get_node(fname, 'spam')
        assert 'x' not in spam.egg
        spam.update_node(egg='x', sausage='Y')
        # Test good node
        save_node(fname, spam, 'spam')
        with open(fname, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        assert 'x' in flags.spam.egg
        # Test bad node
        with pytest.raises(NodeNotFoundError):
            save_node(fname, spam, 'limburger')
        save_node(fname, spam, 'Cheddar', create_new=True)
        with open(fname, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        assert 'Cheddar' in flags
        # Test no preexisting file
        new_fname = Path(tmp_path, 'sample_new.json')
        assert new_fname.exists() is False
        with pytest.raises(FileNotFoundError):
            save_node(new_fname, spam, 'spam')
        # Test new empty file
        empty_fname = Path(tmp_path, 'sample_empty.json')
        empty_fname.touch()
        save_node(empty_fname, spam, 'spam', create_new=True)
        with open(empty_fname, 'r') as f:
            flags = BottleFlags.fromJSON(f.read())
        assert 'spam' in flags



"""Library to create SBE .btl equivalent files.

Joseph Gum SIO/ODF
Nov 7, 2016
"""

import csv
import logging
from collections import OrderedDict
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

from . import flagging as flagging
from . import get_ctdcal_config
from . import oxy_fitting as oxy_fitting

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

BOTTLE_FIRE_COL = "btl_fire"
BOTTLE_FIRE_NUM_COL = "btl_fire_num"


def retrieveBottleDataFromFile(converted_file):
    """
    Retrieve the bottle data from a converted file.
    """
    converted_df = pd.read_pickle(converted_file)

    return retrieveBottleData(converted_df)


def retrieveBottleData(converted_df):
    """
    Retrieve the bottle data from a dataframe created from a converted file.

    Looks for changes in the BOTTLE_FIRE_COL column, ready to be averaged in making the CTD bottle file.
    """
    if BOTTLE_FIRE_COL in converted_df.columns:
        converted_df[BOTTLE_FIRE_NUM_COL] = (
            (
                (converted_df[BOTTLE_FIRE_COL])
                & (
                    converted_df[BOTTLE_FIRE_COL]
                    != converted_df[BOTTLE_FIRE_COL].shift(1)
                )
            )
            .astype(int)
            .cumsum()
        )
        # converted_df['bottle_fire_num'] = ((converted_df[BOTTLE_FIRE_COL] == False)).astype(int).cumsum()
        return converted_df.loc[converted_df[BOTTLE_FIRE_COL]]
        # return converted_df
    else:
        log.error("Bottle fire column:", BOTTLE_FIRE_COL, "not found")

    return pd.DataFrame()  # empty dataframe


def bottle_mean(btl_df):
    """Compute the mean for each bottle from a dataframe."""
    btl_max = int(btl_df[BOTTLE_FIRE_NUM_COL].tail(n=1))
    i = 1
    output = pd.DataFrame()
    while i <= btl_max:
        output = pd.concat(
            (
                output,
                btl_df[btl_df[BOTTLE_FIRE_NUM_COL] == i]
                .mean()
                .to_frame(name=i)
                .transpose(),
            )
        )
        i += 1
    return output


def bottle_median(btl_df):
    """Compute the median for each bottle from a dataframe."""
    btl_max = int(btl_df[BOTTLE_FIRE_NUM_COL].tail(n=1))
    i = 1
    output = pd.DataFrame()
    while i <= btl_max:
        output = pd.concat(
            (
                output,
                btl_df[btl_df[BOTTLE_FIRE_NUM_COL] == i]
                .median()
                .to_frame(name=i)
                .transpose(),
            )
        )
        i += 1
    return output


def _load_btl_data(btl_file, cols=None):
    """
    Loads "bottle mean" CTD data from .pkl file. Function will return all data unless
    cols is specified (as a list of column names)
    """

    btl_data = pd.read_pickle(btl_file)
    if cols is not None:
        btl_data = btl_data[cols]
    btl_data["SSSCC"] = Path(btl_file).stem.split("_")[0]

    return btl_data


def _load_reft_data(reft_file, index_name="btl_fire_num"):
    """
    Loads reft_file to dataframe and reindexes to match bottle data dataframe
    """
    reft_data = pd.read_csv(reft_file, usecols=["btl_fire_num", "T90", "REFTMP_FLAG_W"])
    reft_data.set_index(index_name)
    reft_data["SSSCC_TEMP"] = Path(reft_file).stem.split("_")[0]
    reft_data["REFTMP"] = reft_data["T90"]

    return reft_data


def _load_salt_data(salt_file, index_name="SAMPNO"):
    """
    Loads salt_file to dataframe and reindexes to match bottle data dataframe
    """
    salt_data = pd.read_csv(
        salt_file, usecols=["SAMPNO", "SALNTY", "BathTEMP", "CRavg"]
    )
    salt_data.set_index(index_name)
    salt_data["SSSCC_SALT"] = Path(salt_file).stem.split("_")[0]
    salt_data.rename(columns={"SAMPNO": "SAMPNO_SALT"}, inplace=True)

    return salt_data


def _add_btl_bottom_data(df, cast, lat_col="LATITUDE", lon_col="LONGITUDE", decimals=4):
    """
    Adds lat/lon, date, and time to dataframe based on the values in the bottom_bottle_details.csv
    """
    cast_details = pd.read_csv(
        # cfg.dirs["logs"] + "cast_details.csv", dtype={"SSSCC": str}
        cfg.dirs["logs"] + "bottom_bottle_details.csv",
        dtype={"SSSCC": str},
    )
    cast_details = cast_details[cast_details["SSSCC"] == cast]
    # df[lat_col] = np.round(cast_details["latitude"].iat[0], decimals)
    # df[lon_col] = np.round(cast_details["longitude"].iat[0], decimals)
    df[lat_col] = cast_details["latitude"].iat[0]
    df[lon_col] = cast_details["longitude"].iat[0]

    ts = pd.to_datetime(cast_details["bottom_time"].iat[0], unit="s")
    date = ts.strftime("%Y%m%d")
    hour = ts.strftime("%H%M")
    df["DATE"] = date
    df["TIME"] = hour
    return df


def load_all_btl_files(ssscc_list, cols=None):
    """
    Load bottle and secondary (e.g. reference temperature, bottle salts, bottle oxygen)
    files for station/cast list and merge into a dataframe.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to load
    cols : list of str, optional
        Subset of columns to load, defaults to loading all

    Returns
    -------
    df_data_all : DataFrame
        Merged dataframe containing all loaded data

    """
    df_data_all = pd.DataFrame()

    for ssscc in ssscc_list:
        log.info("Loading BTL data for station: " + ssscc + "...")
        btl_file = cfg.dirs["bottle"] + ssscc + "_btl_mean.pkl"
        btl_data = _load_btl_data(btl_file, cols)

        ### load REFT data
        reft_file = cfg.dirs["reft"] + ssscc + "_reft.csv"
        try:
            reft_data = _load_reft_data(reft_file)
            if len(reft_data) > 36:
                log.error(f"len(reft_data) > 36 for {ssscc}, check reftmp file")
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFT Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            reft_data = pd.DataFrame(index=btl_data.index, columns=["T90"], dtype=float)
            reft_data["btl_fire_num"] = btl_data["btl_fire_num"].astype(int)
            reft_data["SSSCC_TEMP"] = ssscc

        ### load REFC data
        refc_file = cfg.dirs["salt"] + ssscc + "_salts.csv"
        try:
            refc_data = _load_salt_data(refc_file, index_name="SAMPNO")
            if len(refc_data) > 36:
                log.error(f"len(refc_data) > 36 for {ssscc}, check autosal file")
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFC Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            refc_data = pd.DataFrame(
                index=btl_data.index,
                columns=["CRavg", "BathTEMP", "BTLCOND"],
                dtype=float,
            )
            refc_data["SAMPNO_SALT"] = btl_data["btl_fire_num"].astype(int)

        ### load OXY data
        oxy_file = Path(cfg.dirs["oxygen"] + ssscc)
        try:
            oxy_data, params = oxy_fitting.load_winkler_oxy(oxy_file)
            if len(oxy_data) > 36:
                log.error(f"len(oxy_data) > 36 for {ssscc}, check oxygen file")
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFO Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            oxy_data = pd.DataFrame(
                index=btl_data.index,
                columns=[
                    "FLASKNO",
                    "TITR_VOL",
                    "TITR_TEMP",
                    "DRAW_TEMP",
                    "TITR_TIME",
                    "END_VOLTS",
                ],
                dtype=float,
            )
            oxy_data["STNNO_OXY"] = ssscc[:3]
            oxy_data["CASTNO_OXY"] = ssscc[3:]
            oxy_data["BOTTLENO_OXY"] = btl_data["btl_fire_num"].astype(int)

        ### clean up dataframe
        # Horizontally concat DFs to have all data in one DF
        btl_data = pd.merge(btl_data, reft_data, on="btl_fire_num", how="outer")
        btl_data = pd.merge(
            btl_data,
            refc_data,
            left_on="btl_fire_num",
            right_on="SAMPNO_SALT",
            how="outer",
        )
        btl_data = pd.merge(
            btl_data,
            oxy_data,
            left_on="btl_fire_num",
            right_on="BOTTLENO_OXY",
            how="outer",
        )

        if len(btl_data) > 36:
            log.error(f"len(btl_data) for {ssscc} > 36, check bottle file")

        # Add bottom of cast information (date,time,lat,lon,etc.)
        btl_data = _add_btl_bottom_data(btl_data, ssscc)

        # Merge cast into df_data_all
        try:
            df_data_all = pd.concat([df_data_all, btl_data], sort=False)
        except AssertionError:
            raise AssertionError(
                "Columns of " + ssscc + " do not match those of previous columns"
            )
        # print("* Finished BTL data station: " + ssscc + " *")

    # Drop duplicated columns generated by concatenation
    df_data_all = df_data_all.loc[:, ~df_data_all.columns.duplicated()]

    df_data_all["master_index"] = range(len(df_data_all))

    return df_data_all


def _reft_loader(ssscc, reft_dir=cfg.dirs["reft"]):
    """
    Loads SBE35.cap files and assembles into a dataframe.

    Parameters
    ----------
    ssscc : str
        Station to load .CAP file for
    reft_dir : str, pathlib Path
        Path to the reft folder 

    Returns
    -------
    reftDF : DataFrame
        DataFrame of .CAP file with headers

    """
    # semi-flexible search for reft file (in the form of *ssscc.cap)
    try:
        reft_path = sorted(Path(reft_dir).glob(f"*{ssscc}.cap"))[0]
    except IndexError:
        raise FileNotFoundError

    # this works better than pd.read_csv as format is semi-inconsistent (cf .cap files)
    with open(reft_path, "r", newline="") as f:
        reftF = csv.reader(
            f, delimiter=" ", quoting=csv.QUOTE_NONE, skipinitialspace="True"
        )
        reftArray = []
        for row in reftF:
            if len(row) != 17:  # skip over 'bad' rows (empty lines, comments, etc.)
                if len(row) > 17:
                    log.warning(f"Raw REFT for {ssscc} includes line with abnormally large number of columns. Check .CAP file.")
                else:
                    continue
            reftArray.append(row)
    if len(reftArray)>36:
        log.warning("Raw REFT file for {ssscc} exceeds 36 entries. Check .CAP file.")

    reftDF = pd.DataFrame.from_records(reftArray)
    pd.set_option('future.no_silent_downcasting', True) #   Opt in to future downcasting
    reftDF = reftDF.replace(
        to_replace=["bn", "diff", "val", "t90", "="], value=np.nan
    )
    reftDF = reftDF.dropna(axis=1)
    reftDF.loc[:, 1] = reftDF[[1, 2, 3, 4]].agg(" ".join, axis=1)  # dd/mm/yy/time cols are
    reftDF = reftDF.drop(columns=[2, 3, 4])  # read separately; combine into one

    if len(reftDF.columns) != 6:
        log.warning("Raw REFT file for {ssscc} has the incorrect number of columns. Check .CAP file.")
        #   Code will break below

    columns = OrderedDict(  # having this as a dict streamlines next steps
        [
            ("index_memory", int),
            ("datetime", object),
            ("btl_fire_num", int),
            ("diff", int),
            ("raw_value", float),
            ("T90", float),
        ]
    )
    reftDF.columns = list(columns.keys())  # name columns
    reftDF = reftDF.astype(columns)  # force dtypes
    reftDF["datetime"] = pd.to_datetime(reftDF['datetime'], format='%d %b %Y %H:%M:%S') #   Datetime checks, otherwise col is not used

    #   Check contents of the file for stuff that the analyst should double-check (don't immediately flag)
    if any(((reftDF['datetime'] < pd.Timestamp('1985-01-01')) | (reftDF['datetime'] > pd.Timestamp.today()))):
        #   Check timestamps
        log.warning(f"Raw REFT file for {ssscc} contains erroneous timestamps. Check instrument configuration.")
    if any((reftDF['raw_value'] < 50000.0) | (reftDF['raw_value'] > 1000000)):
        #   Check frequencies
        log.warning(f"Raw REFT file reports potentially erroneous frequencies for SSSCC {ssscc}.")
    if any((reftDF['T90'] < -2) | (reftDF['T90'] > 40)):
        #   Check calibration coeffs based on frequencies
        log.warning(f"Raw REFT file reports potentially erroneous temperatures for SSSCC {ssscc}. Check instrument frequencies and calibration coefficients.")
    if any((reftDF['btl_fire_num'] == 1) & (reftDF['btl_fire_num'].shift(1) > 1)):
        #   Check bottle column incrementations
        log.warning(f"Raw REFT file has bottle reset to 1 during {ssscc}. Check .CAP file for tests or other casts.")
    elif reftDF["btl_fire_num"].duplicated().any():
        #   Give warnings if there are any duplicates (file not broken up or deck test bottles not removed)
        log.warning(f"Raw REFT for {ssscc} contains duplicate bottle numbers. Check .CAP file.")

    # assign initial flags (large "diff" = unstable reading, flag questionable)
    reftDF["REFTMP_FLAG_W"] = 2
    reftDF.loc[reftDF["diff"].abs() >= 3000, "REFTMP_FLAG_W"] = 3

    if any(reftDF["REFTMP_FLAG_W"] > 2):    #   Tell the user iteratively in the logs
        for idx, row in reftDF.loc[reftDF["REFTMP_FLAG_W"] == 3].iterrows():
            bad_point = row["index_memory"]
            log.info(f"Measurement {bad_point} flagged questionable in SSSCC {ssscc}")

    # add in STNNBR, CASTNO columns
    # string prob better for other sta/cast formats (names, letters, etc.)
    if len(ssscc) > 5:
        log.warning(f"Length of {ssscc} name exceeds 5. Assigning STNNBR to {ssscc[0:3]}, CASTNO to {ssscc[3:5]}")
    reftDF["STNNBR"] = ssscc[0:3]
    reftDF["CASTNO"] = ssscc[3:5]
    return reftDF


def process_reft(ssscc_list, reft_dir=cfg.dirs["reft"]):
    """
    SBE35 reference thermometer processing function. Load in .cap files for given
    station/cast list, perform basic flagging, and export to .csv files.

    Parameters
    -------
    ssscc_list : list of str
        List of stations to process
    reft_dir : str, optional
        Path to folder containing raw salt files (defaults to data/reft/)

    """
    for ssscc in ssscc_list:
        if not Path(reft_dir + ssscc + "_reft.csv").exists():
            try:
                reftDF = _reft_loader(ssscc, reft_dir)
                reftDF.to_csv(reft_dir + ssscc + "_reft.csv", index=False)
            except FileNotFoundError:
                log.warning(
                    "refT file for cast " + ssscc + " does not exist... skipping"
                )
                continue


def add_btlnbr_cols(df, btl_num_col):
    """
    Initialize bottle number column and initialize WOCE bottle flags.

    Parameters
    ----------
    df : Pandas DataFrame
        Bottle DataFrame containing a defined rosette bottle number
    btl_num_col : String
        String of bottle column to be reassigned

    Returns
    -------
    df : Pandas DataFrame
        Bottle DataFrame with BTLNBR and flag columns as type int
    """
    df["BTLNBR"] = df[btl_num_col].astype(int)
    # default to everything being good
    df["BTLNBR_FLAG_W"] = 2
    return df


def load_hy_file(path_to_hyfile):
    """
    Read in an exchange-formatted bottle file as a Pandas DataFrame.

    Inputs
    ------
    path_to_hyfile : String or Path object
        The path to the bottle file.

    Returns
    -------
    df : Pandas DataFrame
        The bottle file without the lead/end rows, comments, or units
    """

    df = pd.read_csv(path_to_hyfile, comment="#", skiprows=[0])
    df = df.drop(df.index[0])  #   Drop the units
    df = df[df["EXPOCODE"] != "END_DATA"]  #   Drop the final row
    return df


def merge_hy1(
    df1,
    df2,
):
    """
    Merges two hy1 files, returning the combined Pandas DataFrame.
    If the hy1 file has not been loaded yet, use load_hy_file.

    Inputs
    -------
    df1 : Pandas DataFrame
        First hy1 file for concatination
    df2 : Pandas DataFrame
        Second hy1 file for concatination

    Returns
    df: Pandas DataFrame
        Merged bottle file as a DataFrame
    """

    if set(df1.columns) != set(df2.columns):
        print("Bottle file columns do not match. Concatenating with NaNs.")

    df = pd.concat([df1, df2], axis=0, ignore_index=True)  #   Staple df2 onto there

    sorting_cols = {"STNNBR", "CASTNO", "SAMPNO"}
    if sorting_cols.issubset(df):
        if df[list(sorting_cols)].isna().any().any():
            print("NaNs found in station/cast/sample number. Check source files.")

        else:
            df = df.sort_values(
                by=["STNNBR", "CASTNO", "SAMPNO"],
                ascending=[True, True, False],
                ignore_index=True,
            )
    return df


def export_report_data(df):
    """
    Write out the data used for report generation as a csv.

    Params
    ------
    df : Pandas DataFrame
        Fit bottle data

    """
    df["STNNBR"] = [int(x[0:3]) for x in df["SSSCC"]]
    df["CTDPRS"] = df["CTDPRS"].round(1)
    cruise_report_cols = [
        "STNNBR",
        "CTDPRS",
        "CTDTMP1",
        "CTDTMP1_FLAG_W",
        "CTDTMP2",
        "CTDTMP2_FLAG_W",
        "REFTMP",
        "CTDCOND1",
        "CTDCOND1_FLAG_W",
        "CTDCOND2",
        "CTDCOND2_FLAG_W",
        "BTLCOND",
        "CTDSAL",
        "CTDSAL_FLAG_W",
        "SALNTY",
        "CTDOXY",
        "CTDOXY_FLAG_W",
        "CTDRINKO",
        "CTDRINKO_FLAG_W",
        "OXYGEN",
    ]

    # add in missing flags
    df["CTDTMP1_FLAG_W"] = flagging.by_residual(
        df["CTDTMP1"], df["REFTMP"], df["CTDPRS"]
    )
    df["CTDTMP2_FLAG_W"] = flagging.by_residual(
        df["CTDTMP1"], df["REFTMP"], df["CTDPRS"]
    )
    df["CTDCOND1_FLAG_W"] = flagging.by_residual(
        df["CTDCOND1"], df["BTLCOND"], df["CTDPRS"]
    )
    df["CTDCOND2_FLAG_W"] = flagging.by_residual(
        df["CTDCOND2"], df["BTLCOND"], df["CTDPRS"]
    )
    df["CTDOXY_FLAG_W"] = flagging.by_percent_diff(df["CTDOXY"], df["OXYGEN"])
    df["CTDRINKO_FLAG_W"] = flagging.by_percent_diff(df["CTDRINKO"], df["OXYGEN"])

    df[cruise_report_cols].to_csv("data/report_data.csv", index=False)

    return


def export_hy1(df, out_dir=cfg.dirs["pressure"], org="ODF"):
    """
    Write out the exchange-lite formatted hy1 bottle file.

    Params
    ------
    df : Pandas DataFrame
        Fit bottle data
    out_dir = String or Path object, optional
        The path for where to write the hy1 file
    org : String, optional
        The organization or group used to determine subroutines

    """
    log.info("Exporting bottle file")
    btl_data = df.copy()
    now = datetime.now()
    file_datetime = now.strftime("%Y%m%d")

    btl_columns = {
        "EXPOCODE": "",
        "SECT_ID": "",
        "STNNBR": "",
        "CASTNO": "",
        "SAMPNO": "",
        "BTLNBR": "",
        "BTLNBR_FLAG_W": "",
        "DATE": "",
        "TIME": "",
        "LATITUDE": "",
        "LONGITUDE": "",
        "DEPTH": "METERS",
        "CTDPRS": "DBAR",
        "CTDTMP": "ITS-90",
        "CTDSAL": "PSS-78",
        "CTDSAL_FLAG_W": "",
        "SALNTY": "PSS-78",
        "SALNTY_FLAG_W": "",
        # "CTDOXY": "UMOL/KG",
        # "CTDOXY_FLAG_W": "",
        # "CTDRINKO": "UMOL/KG",
        # "CTDRINKO_FLAG_W": "",
        "CTDOXY": "UMOL/KG",
        "CTDOXY_FLAG_W": "",
        "OXYGEN": "UMOL/KG",
        "OXYGEN_FLAG_W": "",
        "REFTMP": "ITS-90",
        "REFTMP_FLAG_W": "",
    }

    # rename outputs as defined in user_settings.yaml
    for param, attrs in cfg.ctd_outputs.items():
        if param not in btl_data.columns:
            btl_data.rename(columns={attrs["sensor"]: param}, inplace=True)

    btl_data["EXPOCODE"] = cfg.expocode
    btl_data["SECT_ID"] = cfg.section_id
    btl_data["STNNBR"] = [int(x[0:3]) for x in btl_data["SSSCC"]]
    btl_data["CASTNO"] = [int(x[3:]) for x in btl_data["SSSCC"]]
    btl_data["SAMPNO"] = btl_data["btl_fire_num"].astype(int)
    btl_data = add_btlnbr_cols(btl_data, btl_num_col="btl_fire_num")

    # sort by decreasing sample number (increasing pressure) and reindex
    btl_data = btl_data.sort_values(
        by=["STNNBR", "SAMPNO"], ascending=[True, False], ignore_index=True
    )

    # switch oxygen primary sensor to rinko
    btl_data["CTDOXY"] = btl_data.loc[:, "CTDRINKO"]
    btl_data["CTDOXY_FLAG_W"] = btl_data.loc[:, "CTDRINKO_FLAG_W"]

    # round data
    # for col in ["CTDTMP", "CTDSAL", "SALNTY", "REFTMP"]:
    #     btl_data[col] = btl_data[col].round(4)
    # for col in ["CTDPRS", "CTDOXY", "OXYGEN"]:
    #     btl_data[col] = btl_data[col].round(1)

    # add depth
    depth_df = pd.read_csv(
        cfg.dirs["logs"] + "depth_log.csv", dtype={"SSSCC": str}, na_values=-999
    ).dropna()
    manual_depth_df = pd.read_csv(
        cfg.dirs["logs"] + "manual_depth_log.csv", dtype={"SSSCC": str}
    )
    full_depth_df = pd.concat([depth_df, manual_depth_df])
    full_depth_df.drop_duplicates(subset="SSSCC", keep="first", inplace=True)
    btl_data["DEPTH"] = -999
    for index, row in full_depth_df.iterrows():
        btl_data.loc[btl_data["SSSCC"] == row["SSSCC"], "DEPTH"] = int(row["DEPTH"])

    # deal with nans
    btl_data["REFTMP_FLAG_W"] = flagging.nan_values(
        btl_data["REFTMP_FLAG_W"], old_flags=btl_data["REFTMP_FLAG_W"]
    )
    btl_data = btl_data.where(~btl_data.isnull(), -999)

    # check columns
    try:
        btl_data[btl_columns.keys()]
        # this is lazy, do better
    except KeyError as err:
        log.info("Column names not configured properly... attempting to correct")
        bad_cols = err.args[0].split("'")[1::2]  # every other str is a column name
        for col in bad_cols:
            if col.endswith("FLAG_W"):
                log.warning(col + " missing, flagging with 9s")
                btl_data[col] = 9
            else:
                log.warning(col + " missing, filling with -999s")
                btl_data[col] = -999

    btl_data = btl_data[btl_columns.keys()]
    time_stamp = file_datetime + org
    with open(out_dir + cfg.expocode + "_hy1.csv", mode="w+") as f:
        f.write("BOTTLE, %s\n" % (time_stamp))
        f.write(",".join(btl_columns.keys()) + "\n")
        f.write(",".join(btl_columns.values()) + "\n")
        btl_data.to_csv(f, header=False, index=False)
        f.write("\n" + "END_DATA")

    return



"""
Unit tests for ctdcal.processors.cast_tools
"""
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from ctdcal.processors.cast_tools import Cast


class TestCast:
    @pytest.fixture
    def tmp_df(self, tmp_path):
        data = {'spam': [1, 2, 3, 2, 1, 2, 3, 4, 5, 4, 3, 2, 1]}
        return pd.DataFrame(data=data)

    @pytest.fixture
    def rnd_df(self):
        cols = ['scan_datetime', 'P', 'GPSLAT', 'GPSLON', 'ALT']
        data = np.random.randint(10, size=[4, 5])
        return pd.DataFrame(data, columns=cols)

    def test_load_cast(self, tmp_df):
        # test happy path
        with patch('pandas.read_pickle', return_value=tmp_df):
            cast = Cast('cast', 'fake_path')
        assert type(cast.proc) == pd.DataFrame
        # test sad path
        with pytest.raises(FileNotFoundError):
            Cast('spam', 'fake_path')

    def test_parse_downcast(self, tmp_df):
        with patch('pandas.read_pickle', return_value=tmp_df):
            cast = Cast('cast', 'fake_dir')
        # test p_col not set
        with pytest.raises(AttributeError):
            cast.parse_downcast('fake_data')
        # test happy path
        cast.p_col = 'spam'
        assert cast.downcast is None
        cast.parse_downcast(cast.proc)
        assert type(cast.downcast) == pd.DataFrame
        assert cast.downcast is not None
        assert cast.downcast['spam'].values[-1] == 5

    def test_filter(self, tmp_df):
        with patch('pandas.read_pickle', return_value=tmp_df):
            cast = Cast('spam', 'fake_dir')
        cast.filter(cast.proc, 2, cols=['spam'])
        assert type(cast.filtered) == pd.DataFrame
        # test bad filter type
        with pytest.raises(AttributeError):
            cast.filter(cast.proc, 2, win_type='egg')

    def test_trim_soak(self, tmp_df):
        with patch('pandas.read_pickle', return_value=tmp_df):
            cast = Cast('spam', 'fake_dir')
        # test p_col not set
        with pytest.raises(AttributeError):
            cast.trim_soak('fake_data', 'fake_win')
        # test downcast not parsed
        cast.p_col = 'spam'
        with pytest.raises(AttributeError):
            cast.trim_soak('fake_data', 'fake_win')
        # test happy path
        cast.parse_downcast(cast.proc)
        cast.trim_soak(cast.downcast, 3, max_soak=2)
        assert type(cast.trimmed) == pd.DataFrame

    def test_get_details(self, rnd_df):
        with patch('pandas.read_pickle', return_value=rnd_df):
            cast = Cast('spam', 'fake_dir')
            # test p_col not set
            with pytest.raises(AttributeError):
                cast.get_details()
        # test downcast not parsed and trimmed
        cast.p_col = 'P'
        with pytest.raises(AttributeError):
            cast.get_details()
        # test happy path
        cast.trimmed = cast.proc
        spam = cast.get_details()
        assert spam.size == 9
        # test no altimeter
        cast.proc.drop('ALT', axis=1)
        cast.trimmed = cast.proc
        spam = cast.get_details()
        assert 'ALT' not in spam.columns
        # test incomplete gps
        cast.proc.drop('GPSLAT', axis=1)
        spam = cast.get_details()
        assert 'GPSLON' not in spam.columns

    # def test_get_pressure_offsets(self):
    #     # TODO: write this when get_pressure_offsets is updated
    #     assert False



import logging
from importlib import resources
from pathlib import Path

import click
from bokeh.application import Application
from bokeh.application.handlers.script import ScriptHandler
from bokeh.server.server import Server
from tornado.ioloop import IOLoop

from . import get_ctdcal_config

# Rich handling
# from rich.logging import RichHandler
# from rich.console import Console
# handler = logging.StreamHandler()
# handler.addFilter(logging.Filter("ctdcal"))  # filter out msgs from other modules
# FORMAT = "%(funcName)s: %(message)s"
# logging.basicConfig(
#     level="INFO",
#     format=FORMAT,
#     datefmt="[%X]",
#     handlers=[handler, RichHandler(console=Console(stderr=True))],
# )

# log = logging.getLogger(__name__)

## logging settings
# terminal output
stream = logging.StreamHandler()
stream.setLevel(logging.WARNING)
stream.addFilter(logging.Filter("ctdcal"))  # filter out msgs from other modules

# ctdcal.log output
logfile_FORMAT = "%(asctime)s | %(funcName)s |  %(levelname)s: %(message)s"
logfile = logging.FileHandler("ctdcal.log")
logfile.setLevel(logging.NOTSET)
logfile.addFilter(logging.Filter("ctdcal"))  # filter out msgs from other modules
logfile.setFormatter(logging.Formatter(logfile_FORMAT))

# global configs
FORMAT = "%(funcName)s: %(levelname)s: %(message)s"
logging.basicConfig(
    level=logging.NOTSET,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[stream, logfile],
)

log = logging.getLogger(__name__)
cfg = get_ctdcal_config()


@click.group()
@click.option("--debug/--no-debug", default=False)
def cli(debug):
    """The ctdcal command creates and manipulates data directories

    Documentation: tbd
    """
    if debug:
        click.echo("Debug mode on (displaying all levels)")
        logging.getLogger("ctdcal").setLevel(logging.NOTSET)
    else:
        click.echo("Debug mode off (displaying 'WARNING' and higher levels)")


@cli.command()
def init():
    """Setup data folder with appropriate subfolders"""

    log.info(f"Building default /data/ directories: \n {*cfg.dirs.keys(),}")

    for sub_dir in cfg.dirs.values():
        Path(sub_dir).mkdir(parents=True)


@cli.command("import")  # click workaround to get a command named 'import'
def import_data():
    """Import data from given folder into ctdcal for processing"""
    # something like this?
    # ctdcal import _path_

    raise NotImplementedError


@cli.command()
@click.option(
    "-g",
    "--group",
    type=click.Choice(["ODF", "PMEL"], case_sensitive=False),
    default="ODF",
)
# @click.option(
#     "-t",
#     "--type",
#     type=click.Choice(["bottle", "ctd", "all"], case_sensitive=False),
#     default="all",
# )
def process(group):
    """Process data using a particular group's methodology"""

    if group == "ODF":
        from .scripts.odf_process_all import odf_process_all

        log.info("Starting ODF processing run")
        odf_process_all()
    elif group == "PMEL":
        # pmel_process()
        raise NotImplementedError


@cli.command()
def cruise_report():
    """Generate bottle residual figures for cruise report"""

    from .scripts.cruise_report import cruise_report_residuals

    cruise_report_residuals()


@cli.command()
def qc():  # pragma: no cover
    """Launch interactive data flagging web app for QA/QC"""
    io_loop = IOLoop.current()
    tools_dir = resources.files("ctdcal.tools")
    with tools_dir.joinpath("data_qc.py") as fname:
        bokeh_app = Application(ScriptHandler(filename=fname))
    server = Server(bokeh_app, io_loop=io_loop)
    server.start()
    server.show("/")
    io_loop.start()


@click.option(
    "-f",
    "--file",
    type=click.Choice(["hy1", "ct1", "all"], case_sensitive=False),
    default="all",
)
@cli.command()
def quick_convert(file):
    """Convert Sea-Bird .cnv files to Exchange CTD (ct1) and bottle (hy1) files."""
    from .scripts.quick_convert import cnv_to_ct1

    if file == "ct1":
        cnv_to_ct1()
    elif file == "hy1":
        # cnv_to_hy1()
        raise NotImplementedError
    elif file == "all":
        cnv_to_ct1()
        # cnv_to_hy1()

    # should we have non-Exchange output options?


if __name__ == "__main__":
    cli()



"""
A module for handling flags for bottle and continuous data.
"""

import numpy as np
import pandas as pd


def _merge_flags(new_flags, old_flags, keep_higher=True):
    """
    Merge old and new flags into one flag vector.

    Returns new_flags if old_flags is None.

    Parameters
    ----------
    new_flags : array-like
        Newly calculated flags
    old_flags : int, optional
        User-supplied original data flags
    keep_higher : bool, optional
        Whether to keep higher (default) or lower value flags

    Returns
    -------
    merged_flags : array-like
        Merged old and new flags
    """

    if old_flags is None:
        return new_flags

    new_flags = np.squeeze(new_flags)
    old_flags = np.squeeze(old_flags)
    merged_flags = np.copy(new_flags)

    if keep_higher:
        is_higher = old_flags > new_flags
        merged_flags[is_higher] = old_flags[is_higher]
    else:
        # can't really think of a use case for this... maybe delete?
        is_lower = old_flags < new_flags
        merged_flags[is_lower] = old_flags[is_lower]

    return merged_flags


def nan_values(data, old_flags=None, flag_good=2, flag_nan=9):
    """
    Flag values as either good or nan.

    Parameters
    ----------
    data : array-like
        Variable to be flagged
    old_flags : array-like, optional
        Original data flags to be merged in (if provided)
    flag_good : int, optional
        Flag value for good data
    flag_nan : int, optional
        Flag value for bad (nan) data

    Returns
    -------
    flags : array-like
        Flag for each data point in input
    """

    data = np.squeeze(data)
    flags = np.full(np.shape(data), flag_good)
    flags[np.isnan(data)] = flag_nan

    return _merge_flags(flags, old_flags)


def outliers(
    data,
    old_flags=None,
    flag_good=2,
    flag_outlier=4,
    n_sigma1=2,
    n_sigma2=20,
    ignore_nan=True,
):
    """
    Flag extreme outliers using standard deviations from the mean as a threshold.

    Outliers are identified over two passes. For the first pass, mean and standard
    deviation of data are calculated for all data. Values more than n_sigma1 standard
    deviations from mean are (temporarily) flagged questionable. For the second pass,
    mean and standard deviation are re-calculated with questionable data excluded. Data
    more than n_sigma2 standard deviations from mean are flagged as outliers.

    Parameters
    ----------
    data : array-like
        Variable to be flagged
    old_flags : array-like, optional
        Original data flags to be merged in (if provided)
    flag_good : int, optional
        Flag value for good data
    flag_outlier : int, optional
        Flag value for outliers
    n_sigma1 : int, optional
        Number of standard deviations away from mean needed to be excluded from statistics
    n_sigma2 : int, optional
        Number of standard deviations away from mean needed to be outlier
    ignore_nan : bool, optional
        Ignore nan values in data

    Returns
    -------
    flags : array-like
        Flag for each data point in input

    Notes
    -----
    Functionality is similar to Sea-Bird's "Wild Edit" in Seasoft V2.
    """

    data = np.squeeze(data)

    # np.squeeze converts DataFrame->Series but does nothing to Series
    if type(data) is pd.Series:
        data = data.to_numpy()  # force to np.array

    flags = np.full(np.shape(data), flag_good).squeeze()

    # function aliases
    if ignore_nan:
        mean, std = np.nanmean, np.nanstd
    else:
        mean, std = np.mean, np.std

    # pass 1
    questionable = np.abs(data - mean(data)) > (n_sigma1 * std(data))

    # pass 2
    data_mean = mean(data[~questionable])
    data_std = std(data[~questionable])
    outliers = np.abs(data - data_mean) > (n_sigma2 * data_std)
    flags[outliers] = flag_outlier

    return _merge_flags(flags, old_flags)


def by_percent_diff(
    data, ref_data, old_flags=None, percent_thresh=1, flag_good=2, flag_bad=3
):
    """
    Flag residuals based on the percent error relative to the reference data.

    Parameters
    ----------
    data : array-like
        Variable to be flagged
    ref_data : array-like
        Reference data to compare against
    old_flags : array-like, optional
        Original data flags to be merged in (if provided)
    percent_thresh : int, optional
        Percent different cutoff to be flagged as bad data
    flag_good : int, optional
        Flag value for good data
    flag_bad : int, optional
        Flag value for bad data

    Returns
    -------
    flags : array-like
        Flag for each data point in input
    """
    data = np.squeeze(data)
    ref_data = np.squeeze(ref_data)

    #   Avoid dividing by zero whenever possible
    data = np.where(data == 0, np.nan, data)

    percent_diff = (np.abs(data - ref_data) / data).squeeze() * 100
    flags = np.full(np.shape(percent_diff), flag_good).squeeze()

    flags[percent_diff > percent_thresh] = flag_bad

    return _merge_flags(flags, old_flags)


def by_residual(
    data,
    ref_data,
    pressure,
    old_flags=None,
    flag_good=2,
    flag_bad=3,
    threshold=[0.002, 0.005, 0.01, 0.02],
    p_cutoff=[2000, 1000, 500],
):
    """
    Flag residuals using specific thresholds for each pressure range.

    Parameters
    ----------
    data : array-like
        Variable to be flagged
    ref_data : array-like
        Reference data to compare against
    flag_good : int, optional
        Flag value for good data
    flag_bad : int, optional
        Flag value for data outside threshold
    threshold : list of float, optional
        Threshold between good and bad values
    p_cutoff : list of int, optional
        Edges of pressure bins

    Returns
    -------
    flags : array-like
        Flag for each data point in input
    """
    if (len(threshold) - len(p_cutoff)) != 1:
        raise IndexError("threshold must be one value longer than p_cutoff")

    if any(np.diff(p_cutoff) >= 0):
        raise ValueError("p_cutoff must be monotonically decreasing")

    residual = np.abs(data - ref_data).squeeze()
    flags = np.full(np.shape(residual), flag_good).squeeze()

    # first and last pressure bin need special handling
    is_deep = pressure > p_cutoff[0]
    is_shallow = pressure < p_cutoff[-1]
    flags[is_deep & (residual > threshold[0])] = flag_bad
    flags[is_shallow & (residual > threshold[-1])] = flag_bad

    p_bins = zip(p_cutoff, p_cutoff[1:])  # make pairs (e.g. (2000, 1000), (1000, 500))
    threshold = threshold[1:-1]  # deep and shallow threshold values
    for (p_upper, p_lower), thresh in zip(p_bins, threshold):
        in_bin = (pressure <= p_upper) & (pressure > p_lower)
        flags[in_bin & (residual > thresh)] = flag_bad

    return _merge_flags(flags, old_flags)



import io
import logging
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd

from ctdcal import odf_io


def check_type(to_check, sub_dtype):
    """
    Type comparisons can fail on Windows due to int32/int64 behavior.
    This function checks subtypes so that tests don't fail on Windows.

    See https://stackoverflow.com/q/64901822/13155619 for more info
    """
    if sub_dtype is int:
        np_type = np.integer
    elif sub_dtype is float:
        np_type = np.floating

    return all([np.issubdtype(dtype, np_type) for dtype in to_check.dtypes])


def make_salt_file(stn=1, cast=1, comment=None, flag=False, to_file=None):
    #  seed RNG for Reading3, comments, and flags
    rng = np.random.default_rng(seed=100)

    # build dummy DataFrame
    constants = {
        "STNNBR": f"{stn:04.0f}",
        "CASTNO": f"{cast:02.0f}",
        "BathTemp": 21,
        "unk": 5907,
    }
    salts = pd.DataFrame(data=constants, index=np.arange(12))
    salts.insert(2, "SAMPNO", [f"{x:02.0f}" for x in salts.index])
    salts.insert(4, "CRavg", np.linspace(1.95, 1.98, 12)[::-1])
    salts.insert(5, "autosalSAMPNO", salts["SAMPNO"].astype(int))
    salts["autosalSAMPNO"] = salts["autosalSAMPNO"].astype(
        object
    )  #   Change dtype for integers and strings
    salts.loc[[0, 11], "autosalSAMPNO"] = "worm"
    times = pd.date_range(start="18:32:04", end="19:36:03", periods=24)
    salts["StartTime"] = times[::2].strftime("%H:%M:%S")
    salts["EndTime"] = times[1::2].strftime("%H:%M:%S")

    # assign CR values around mean (give only some samples 3 readings)
    salts["Reading1"] = salts["CRavg"] + 1e-4
    salts["Reading2"] = salts["CRavg"] - 1e-4
    salts["Reading3"] = salts["CRavg"] * rng.choice([1, np.nan], 12)
    attempts = salts[["Reading1", "Reading2", "Reading3"]].count(axis=1)
    salts.insert(9, "Attempts", attempts.map("{:02.0f}".format))

    # add comment marker (#, x)
    if comment is not None:
        salts["STNNBR"] = rng.choice(["", comment], 12, p=[0.8, 0.2]) + salts["STNNBR"]

    # add LabView "quality"(?) flag
    if flag:
        salts["EndTime"] = salts["EndTime"] + rng.choice(["", "*"], 12, p=[0.8, 0.2])

    # final formatting, remove blank Reading3 cells to match Autosal
    header = (
        "12-345 operator: ABC box: S batch: P678 k15: 0.91011 std dial 408 01/01/2000"
    )
    string_df = salts.to_string(
        header=False, index=False, float_format="{:.5f}".format, na_rep=""
    )
    text_out = "\n".join([header, string_df.replace("        ", "")])

    if to_file is not None:
        with open(to_file, "w+") as f:
            f.write(text_out)
    else:
        return text_out


def test_salt_loader(caplog, tmp_path):
    # check salt file loads in correctly
    salt_file = make_salt_file()
    saltDF, refDF, q = odf_io._salt_loader(io.StringIO(salt_file))

    assert saltDF.shape == (10, 14)
    assert all(saltDF[["StartTime", "EndTime"]].dtypes == object)
    assert check_type(saltDF[["CRavg", "Reading1", "Reading2", "Reading3"]], float)
    assert check_type(saltDF[["STNNBR", "CASTNO", "SAMPNO", "autosalSAMPNO"]], int)
    assert check_type(saltDF[["BathTEMP", "Unknown", "Attempts", "IndexTime"]], int)
    assert all(saltDF.index == np.arange(1, 11))
    assert saltDF["Reading3"].isna().sum() == 5
    assert q is None

    assert refDF.shape == (2, 2)
    assert all(refDF.dtypes == float)
    assert all(refDF.index == [0, 11])

    # check commented lines are ignored ("bottles" 1, 4, 10)
    salt_file = make_salt_file(comment="#")
    with caplog.at_level(logging.DEBUG):
        saltDF, refDF, q = odf_io._salt_loader(io.StringIO(salt_file))
        assert "(#, x)" in caplog.messages[0]
        assert "test_odf_io" in caplog.messages[0]
    assert saltDF.shape == (7, 14)
    assert all(saltDF[["StartTime", "EndTime"]].dtypes == object)
    assert check_type(saltDF[["CRavg", "Reading1", "Reading2", "Reading3"]], float)
    assert check_type(saltDF[["STNNBR", "CASTNO", "SAMPNO", "autosalSAMPNO"]], int)
    assert check_type(saltDF[["BathTEMP", "Unknown", "Attempts", "IndexTime"]], int)
    assert all(saltDF.index == [2, 3, 5, 6, 7, 8, 9])
    assert saltDF["Reading3"].isna().sum() == 3

    assert refDF.shape == (2, 2)
    assert all(refDF.dtypes == float)
    assert all(refDF.index == [0, 11])

    # check flagged EndTimes are added to flag_file ("bottles" 1, 4, 10)
    salt_file = make_salt_file(flag=True)
    with caplog.at_level(logging.DEBUG):
        saltDF, refDF, flagged = odf_io._salt_loader(io.StringIO(salt_file))
        assert "Found * in test_odf_io" in caplog.messages[1]
        assert "flagged" in flagged.at[1, "notes"]
        assert len(flagged) == 3
    assert saltDF.shape == (10, 14)
    assert all(saltDF[["StartTime", "EndTime"]].dtypes == object)
    assert check_type(saltDF[["CRavg", "Reading1", "Reading2", "Reading3"]], float)
    assert check_type(saltDF[["STNNBR", "CASTNO", "SAMPNO", "autosalSAMPNO"]], int)
    assert check_type(saltDF[["BathTEMP", "Unknown", "Attempts", "IndexTime"]], int)
    assert all(saltDF.index == np.arange(1, 11))
    assert saltDF["Reading3"].isna().sum() == 5

    assert refDF.shape == (2, 2)
    assert all(refDF.dtypes == float)
    assert all(refDF.index == [0, 11])

    # check behavior with filename (saves time not writing full file to disk 3x)
    d = tmp_path / "salt"
    d.mkdir()
    fake_file = d / "90909"
    fake_file.write_text(
        "12-345 operator: ABC box: S batch: P678 k15: 0.91011 std dial 408 01/01/2000\n1 2 3 4 5 6 7 00:01:00 00:02:00 10 11"
    )
    saltDF, refDF, q = odf_io._salt_loader(fake_file)
    assert all(saltDF[["StartTime", "EndTime"]].dtypes == object)
    assert check_type(saltDF[["CRavg", "Reading1"]], float)
    assert check_type(saltDF[["STNNBR", "CASTNO", "SAMPNO", "autosalSAMPNO"]], int)
    assert check_type(saltDF[["BathTEMP", "Unknown", "Attempts", "IndexTime"]], int)
    assert saltDF.shape == (1, 12)
    assert refDF.empty


def test_remove_autosal_drift(caplog):
    index_time = np.linspace(0, 300, 7)
    CRavg = np.array([2.0] * 6 + [1.4])
    saltDF = pd.DataFrame(data={"STNNBR": 1, "CASTNO": 1}, index=np.arange(1, 6))
    saltDF["IndexTime"], saltDF["CRavg"] = (index_time[1:-1], CRavg[1:-1])
    refDF = pd.DataFrame(data={"IndexTime": index_time[::6], "CRavg": CRavg[::6]})

    # check that drift is removed correctly
    removed = odf_io.remove_autosal_drift(saltDF, refDF)
    np.testing.assert_allclose(removed["CRavg"], np.arange(1.5, 2, 0.1)[::-1])
    assert "IndexTime" not in removed.columns

    # check input DataFrames have not been modified
    assert all(saltDF["CRavg"] == CRavg[1:-1])
    assert all(refDF["CRavg"] == CRavg[::6])

    # return unmodified saltDF if refDF is wrong size
    original = odf_io.remove_autosal_drift(saltDF, refDF.iloc[0])
    assert all(original["CRavg"] == saltDF["CRavg"])
    assert "IndexTime" not in original.columns
    assert "start/end reference" in caplog.messages[0]
    assert "00101" in caplog.messages[0]


def test_salt_exporter(tmp_path, caplog):
    saltDF = pd.DataFrame(data={"STNNBR": 1, "CASTNO": 1}, index=np.arange(5))
    saltDF["CRavg"] = np.ones(5)

    # check test data completes round trip
    assert not Path(tmp_path / "00101_salts.csv").exists()
    odf_io._salt_exporter(saltDF, outdir=str(tmp_path))
    assert Path(tmp_path / "00101_salts.csv").exists()
    with open(Path(tmp_path / "00101_salts.csv")) as f:
        assert saltDF.equals(pd.read_csv(f))

    # check "file already exists" message
    with caplog.at_level(logging.INFO):
        odf_io._salt_exporter(saltDF, outdir=str(tmp_path))
        assert "00101_salts.csv already exists" in caplog.messages[0]

    # check file write with multiple stations
    saltDF["STNNBR"] = [1, 1, 2, 2, 2]
    odf_io._salt_exporter(saltDF, outdir=str(tmp_path))
    assert Path(tmp_path / "00201_salts.csv").exists()

    # check file write with multiple casts
    saltDF["CASTNO"] = [1, 2, 1, 2, 3]
    odf_io._salt_exporter(saltDF, outdir=str(tmp_path))
    assert Path(tmp_path / "00102_salts.csv").exists()
    assert Path(tmp_path / "00202_salts.csv").exists()
    assert Path(tmp_path / "00203_salts.csv").exists()

    # check empty (all NaN) Reading# columns are dropped
    saltDF["STNNBR"], saltDF["CASTNO"] = 3, 1
    saltDF["Reading1"] = 1
    saltDF["Reading2"] = np.nan
    assert not Path(tmp_path / "00301_salts.csv").exists()
    odf_io._salt_exporter(saltDF, outdir=str(tmp_path))
    with open(Path(tmp_path / "00301_salts.csv")) as f:
        empty = pd.read_csv(f)
        assert "Reading1" in empty.columns
        assert "Reading2" not in empty.columns


def test_process_salts(tmp_path, caplog):
    # check missing salt files are logged and skipped
    odf_io.process_salts(["90909"], None, salt_dir=str(tmp_path))
    assert "90909 does not exist" in caplog.messages[0]

    # check salt file is processed through to .csv without errors
    f_path = tmp_path / "90909"
    assert not Path(tmp_path / "90909").exists()
    make_salt_file(stn=909, cast=9, to_file=f_path)  # outfile name depends on stn/cast
    with caplog.at_level(logging.INFO):
        assert not Path(tmp_path / "90909_salts.csv").exists()
        odf_io.process_salts(["90909"], None, salt_dir=str(tmp_path))
        assert len(caplog.messages) == 1
        assert Path(tmp_path / "90909_salts.csv").exists()

    # check nothing happens if .csv already exists
    with caplog.at_level(logging.INFO):
        odf_io.process_salts(["90909"], None, salt_dir=str(tmp_path))
        assert "90909_salts.csv already exists" in caplog.messages[1]

    def test_print_progress_bar():
        # Test parameters
        iteration = 3
        total = 10
        prefix = "Progress"
        suffix = "Complete"
        decimals = 1
        length = 20
        fill = "#"
        printEnd = "\r"

        # Expected output
        expected_output = "\rProgress |###-------| 30.0% Complete"

        # Redirect stdout to mock_stdout
        with patch("sys.stdout", io.StringIO()):
            # Call function
            odf_io.print_progress_bar(
                iteration, total, prefix, suffix, decimals, length, fill, printEnd
            )

        # Get value from mock_stdout
        actual_output = io.StringIO().getvalue()

        # Check if the actual output matches the expected output
        assert actual_output == expected_output



import logging
from pathlib import Path

from ctdcal import flagging, get_ctdcal_config, io

log = logging.getLogger(__name__)

cfg = get_ctdcal_config()


def cnv_to_ct1():
    """
    A script for converting Sea-Bird .cnv files to uncalibrated Exchange CTD files
    for plotting and QA/QC purposes (in ODV, etc.).
    """
    cnv_files = list(Path(cfg.dirs["converted"]).glob("*.cnv"))
    log.info(f"Found {len(cnv_files)} .cnv files, attempting to convert")

    # make CTD files
    sbe_to_woce = {  # map col names from SBE to WOCE
        "prDM": "CTDPRS",
        "t090C": "CTDTMP1",
        "t190C": "CTDTMP2",
        "c0mS/cm": "CTDCOND1",
        "c1mS/cm": "CTDCOND2",
        "sal00": "CTDSAL1",
        "sal11": "CTDSAL2",
        "sbeox0V": "CTDOXYVOLTS",
        "sbeox0ML/L": "CTDOXY",
        "flECO-AFL": "CTDFLUOR",
        "CStarTr0": "CTDXMISS",
    }
    for f in cnv_files:
        df = io.load_cnv(f).rename(mapper=sbe_to_woce, axis=1)
        df = df[sbe_to_woce.values()]

        # give everything WOCE-named uncalibrated flags
        for idx, col in enumerate(df.columns):
            flags = flagging.nan_values(df[col], flag_good=1, flag_nan=9)
            df.insert(idx * 2 + 1, col + "_FLAG_W", flags)

        # export to pressure folder
        df.to_csv(f"{cfg.dirs['pressure']}{f.stem}_ct1.csv", na_rep="-999")


def cnv_to_hy1():
    """
    A script for converting Sea-Bird .cnv files to an uncalibrated Exchange bottle file
    for plotting and QA/QC purposes (in ODV, etc.).
    """
    # this will probably have significant overlap with cnv_to_ct1()
    # maybe there should be three functions in this file?:
    # _load_cnv_files (load/map names, add flags)
    # cnv_to_ct1 (bin to 2db and export file)
    # cnv_to_hy1 (match to data from .bl and export file)

    # something like this to start:

    # bl_files = list(Path(cfg.dirs["converted"]).glob("*.bl"))
    # if not bl_files:
    #     log.warning("Not .bl files found - cannot make Exchange bottle files")

    # find data to average using .bl file
    pass



import numpy as np
import pandas as pd
import pytest
import yaml

from ctdcal import fit_ctd


@pytest.mark.parametrize("xN, yN", [(1, 0), (0, 1), (1, 1), (2, 1), (1, 2)])
def test_multivariate_fit(xN, yN):
    data = [0.0, 0.0]
    coef_names = ["x", "y"]
    x_coefs = [f"x{n}" for n in np.arange(1, xN + 1)] if xN > 0 else []
    y_coefs = [f"y{n}" for n in np.arange(1, yN + 1)] if yN > 0 else []

    # ensure number of coefs produced is correct (including constant offset)
    np.testing.assert_allclose(
        fit_ctd.multivariate_fit(data, (data, xN), (data, yN)), np.zeros(xN + yN + 1)
    )

    # check that coefs are named properly
    fit = fit_ctd.multivariate_fit(data, (data, xN), (data, yN), coef_names=coef_names)
    assert all(coef in fit.keys() for coef in x_coefs + y_coefs)
    assert "c0" in fit.keys()

    # check error if len of inputs and coefs differ
    with pytest.raises(ValueError):
        fit_ctd.multivariate_fit(data, (data, xN), (data, yN), coef_names=["x"])

    # check error if input is not tuple
    with pytest.raises(TypeError):
        fit_ctd.multivariate_fit(data, (data, xN), [data, yN])


def test_apply_polyfit():
    y = np.array([1, 2, 3])

    # check correction is applied correctly (no dependent variables)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0,)), y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (1, 0)), y + 1)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0, 1)), y + y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0, 0.5)), y + 0.5 * y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0, 0, 1)), y + y ** 2)

    # check correction is applied correctly (with dependent variables)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0,), (y, (0,))), y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0,), (y, (1,))), y + y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0,), (y, (1.0,))), y + y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0.0,), (y, (1,))), y + y)
    np.testing.assert_array_equal(fit_ctd.apply_polyfit(y, (0.0,), (y, (1.0,))), y + y)
    np.testing.assert_array_equal(
        fit_ctd.apply_polyfit(y, (0,), (y, (0, 1))), y + y ** 2
    )

    # check error if input is not tuple
    with pytest.raises(TypeError):
        fit_ctd.apply_polyfit(y, (0,), [y, (0,)])

def test_generate_yaml(tmp_path):
    fname = str(tmp_path) + "filename.yaml"
    fit_ctd.generate_yaml("filename.yaml", str(tmp_path))
    with open(fname, 'r') as f:
        generated_data = yaml.safe_load(f)

    assert generated_data['t2']['ssscc_t1']['T_order'] == 0
    assert generated_data['c1']['ssscc_c1']['zRange'] == '1000:6000'



import numpy as np
import pandas as pd
import pytest

from ctdcal import ctd_plots


def test_residual_vs_pressure(tmp_path):
    param = pd.Series(data=[0, 1, 2], name="param")
    ref = pd.Series(data=[0, 1, 2], name="ref")
    prs = pd.Series(data=[0, 3000, 6000], name="prs")
    stn = pd.Series(data=[0, 1, 2], name="stn")

    # check titles, labels, limits, grid
    axes = ctd_plots.residual_vs_pressure(param, ref, prs, stn=stn)
    grid_lines = axes.get_xgridlines() + axes.get_ygridlines()
    assert axes.get_title() == "ref-param vs. prs"
    assert axes.get_xlabel() == "Residual"
    assert axes.get_ylabel() == "Pressure (dbar)"
    assert axes.get_xlim() == (-0.02, 0.02)
    assert axes.get_ylim() == (6000, 0)
    assert axes.collections[0].colorbar is not None
    assert not all(line.get_visible() for line in grid_lines)

    # check titles, labels, limits, grid (with deep setting)
    deep = ctd_plots.residual_vs_pressure(param, ref, prs, stn=stn, deep=True)
    grid_lines = deep.get_xgridlines() + deep.get_ygridlines()
    y_data = deep.collections[0].get_offsets().data[:, 1]
    assert deep.get_title() == "ref-param (>2000 dbar) vs. prs"
    assert deep.get_xlabel() == "Residual"
    assert deep.get_ylabel() == "Pressure (dbar)"
    assert deep.get_xlim() == (-0.02, 0.02)
    assert deep.get_ylim() == (6000, 0)
    assert deep.collections[0].colorbar is not None
    assert not all(line.get_visible() for line in grid_lines)
    assert all(y_data > 2000)

    # check settings are applied properly
    settings = ctd_plots.residual_vs_pressure(
        param,
        ref,
        prs,
        stn=stn,
        xlim=(-1, 1),
        ylim=(2000, 1000),
        xlabel="x",
        ylabel="y",
        auto_title=False,
        grid=True,
    )
    grid_lines = settings.get_xgridlines() + settings.get_ygridlines()
    assert settings.get_xlim() == (-1, 1)
    assert settings.get_ylim() == (2000, 1000)
    assert settings.get_xlabel() == "x"
    assert settings.get_ylabel() == "y"
    assert settings.get_title() == ""
    assert settings.collections[0].colorbar is not None
    assert all(line.get_visible() for line in grid_lines)

    # check behavior with unlabeled data inputs and no stations
    x = np.array([0, 0, 0])
    unlabeled = ctd_plots.residual_vs_pressure(x, x, x)
    assert unlabeled.get_title() == ""
    assert unlabeled.collections[0].colorbar is None

    # check figure saving
    # for ext in [".jpg", ".png", ".pdf"]:
    #     with (tmp_path / "figures" / f"fig1{ext}") as f_out:
    #         assert not f_out.exists()
    #         ctd_plots.residual_vs_pressure(x, x, x, stn=x, f_out=f_out)
    #         assert f_out.exists()

    for ext in [".jpg", ".png", ".pdf"]:
        output_path = tmp_path / "figures"
        output_path.mkdir(parents=True, exist_ok=True)
        file_path = output_path / f"fig1{ext}"

        assert not file_path.exists()
        ctd_plots.residual_vs_pressure(x, x, x, stn=x, f_out=file_path)
        assert file_path.exists()   # Verify that the file has been created


def test_residual_vs_station(tmp_path):
    param = pd.Series(data=[0, 1, 2], name="param")
    ref = pd.Series(data=[0, 1, 2], name="ref")
    prs = pd.Series(data=[0, 3000, 6000], name="prs")
    stn = pd.Series(data=[0, 1, 2], name="stn")

    # check titles, labels, limits
    axes = ctd_plots.residual_vs_station(param, ref, prs, stn)
    grid_lines = axes.get_xgridlines() + axes.get_ygridlines()
    assert axes.get_title() == "ref-param vs. stn"
    assert axes.get_xlabel() == "Station Number"
    assert axes.get_ylabel() == "Residual"
    # xlim intentionally skipped as it's not currently a setting for this function
    assert axes.get_ylim() == (-0.02, 0.02)
    assert not all(line.get_visible() for line in grid_lines)

    # check titles, labels, limits (with deep setting)
    deep = ctd_plots.residual_vs_station(param, ref, prs, stn, deep=True)
    grid_lines = deep.get_xgridlines() + deep.get_ygridlines()
    z_data = deep.collections[0].get_array().data
    assert deep.get_title() == "ref-param (>2000 dbar) vs. stn"
    assert deep.get_xlabel() == "Station Number"
    assert deep.get_ylabel() == "Residual"
    # xlim intentionally skipped as it's not currently a setting for this function
    assert deep.get_ylim() == (-0.02, 0.02)
    assert not all(line.get_visible() for line in grid_lines)
    assert all(z_data > 2000)

    # check settings are applied properly
    settings = ctd_plots.residual_vs_station(
        param,
        ref,
        prs,
        stn,
        ylim=(-1, 1),
        xlabel="x",
        ylabel="y",
        grid=True,
    )
    grid_lines = settings.get_xgridlines() + settings.get_ygridlines()
    assert settings.get_ylim() == (-1, 1)
    assert settings.get_xlabel() == "x"
    assert settings.get_ylabel() == "y"
    assert all(line.get_visible() for line in grid_lines)

    # check behavior with unlabeled data inputs
    x = np.array([0, 0, 0])
    unlabeled = ctd_plots.residual_vs_station(x, x, x, x)
    assert unlabeled.get_title() == ""

    # check figure saving
    for ext in [".jpg", ".png", ".pdf"]:
        # with (tmp_path / "figures" / f"fig2{ext}") as f_out:
        #     assert not f_out.exists()
        #     ctd_plots.residual_vs_station(x, x, x, x, f_out=f_out)
        #     assert f_out.exists()
        output_path = tmp_path / "figures"
        output_path.mkdir(parents=True, exist_ok=True)
        file_path = output_path / f"fig2{ext}"

        assert not file_path.exists()
        ctd_plots.residual_vs_pressure(x, x, x, stn=x, f_out=file_path)
        assert file_path.exists()   # Verify that the file has been created


def test_intermediate_residual_plot(tmp_path):
    param = pd.Series(data=[0, 1, 2], name="param")
    ref = pd.Series(data=[0, 1, 2], name="ref")
    prs = pd.Series(data=[0, 3000, 6000], name="prs")
    stn = pd.Series(data=[0, 1, 2], name="stn")

    # check settings are applied properly
    settings = ctd_plots._intermediate_residual_plot(
        ref - param, prs, stn, xlim=(-1, 1), xlabel="x", show_thresh=True
    )
    grid_lines = settings.get_xgridlines() + settings.get_ygridlines()
    thresh_left = settings.lines[0].get_xydata()
    thresh_right = settings.lines[1].get_xydata()
    assert settings.get_xlim() == (-1, 1)
    assert settings.get_xlabel() == "x"
    assert settings.get_title() == "Mean: 0.0 / Stdev: 0.0"
    assert all(line.get_visible() for line in grid_lines)
    assert all(thresh_left[:, 0] == np.array([0.002, 0.005, 0.010, 0.020, 0.020]))
    assert all(thresh_right[:, 0] == -1 * np.array([0.002, 0.005, 0.010, 0.020, 0.020]))
    assert all(thresh_left[:, 1] == np.array([6000, 2000, 1000, 500, 0]))
    assert all(thresh_right[:, 1] == np.array([6000, 2000, 1000, 500, 0]))

    # check figure saving
    x = np.array([0, 0, 0])
    for ext in [".jpg", ".png", ".pdf"]:
        # with (tmp_path / "figures" / f"fig3{ext}") as f_out:
        #     assert not f_out.exists()
        #     ctd_plots._intermediate_residual_plot(x, x, x, f_out=f_out)
        #     assert f_out.exists()
        output_path = tmp_path / "figures"
        output_path.mkdir(parents=True, exist_ok=True)
        file_path = output_path / f"fig3{ext}"

        assert not file_path.exists()
        ctd_plots.residual_vs_pressure(x, x, x, stn=x, f_out=file_path)
        assert file_path.exists()   # Verify that the file has been created


@pytest.fixture
def example_data():
    return {
        'param1': np.array([33, 33.5, 34, 35, 35.5]),
        'param2': np.array([0.1, 10, 15, 20, 20]),
        'label1': "Salinity",
        'label2': "Temperature",
        'stn': np.array(["1", "2", "3", "4", "5"]),
        'tsS': np.array([33, 33.5, 34, 35, 35.5]),
        'tsT': np.array([0.1, 10, 15, 20, 20]),
    }

def test_param_vs_param(tmp_path, example_data):
    for ext in [".jpg", ".png", ".pdf"]:
        output_path = tmp_path / "figures"
        output_path.mkdir(parents=True, exist_ok=True)
        file_path = output_path / f"not_TS{ext}"

        #   Check that the plotted figure generates and is saved as expected
        assert not file_path.exists()
        ax = ctd_plots.param_vs_param(example_data["param1"], example_data["label1"], 
                                      example_data["param2"], example_data["label2"])   #   Should return figure axis
        assert np.allclose(ax.get_xticks(), [32.5, 33. , 33.5, 34. , 34.5, 35. , 35.5, 36. ])
        assert np.allclose(ax.get_yticks(), [-2.5,  0. ,  2.5,  5. ,  7.5, 10. , 12.5, 15. , 17.5, 20. , 22.5])
        assert ax.get_xlabel() == example_data["label1"]
        assert ax.get_ylabel() == example_data["label2"]
        ctd_plots.param_vs_param(example_data["param1"], example_data["label1"], 
                                 example_data["param2"], example_data["label2"], f_out=file_path)
        assert file_path.exists()

        #   Now check that the TS contour plotting subroutine works
        file_path = output_path / f"TS{ext}"
        assert not file_path.exists()
        ax = ctd_plots.param_vs_param(example_data["param1"], example_data["label1"], 
                                      example_data["param2"], example_data["label2"], 
                                      stn=example_data["stn"], tsT=example_data["tsT"], tsS=example_data["tsS"])
        assert np.allclose(ax.get_xticks(), [32.5, 33. , 33.5, 34. , 34.5, 35. , 35.5, 36. ])  #   Should be unchanged
        assert len(ax.collections[-1].get_paths()) == 9 #   Should generate 9 contours
        assert len(ax.collections[-1].labelTexts) == 7  #   Should have text on 7 of the contour lines
        assert ax.collections[-1].labelTexts[4].get_text() == "26.4"    #   The density calculations should have produced specific text on the 5th contour
        ctd_plots.param_vs_param(example_data["param1"], example_data["label1"], 
                                 example_data["param2"], example_data["label2"], 
                                 f_out=file_path, stn=example_data["stn"], tsT=example_data["tsT"], tsS=example_data["tsS"])
        assert file_path.exists()



import logging
from pathlib import Path

import numpy as np
import scipy

from ctdcal import oxy_fitting


def test_gather_oxy_params(caplog, tmp_path):
    # make fake file path
    f_path = tmp_path / "90909"
    # breakpoint()
    assert not Path(tmp_path / "90909").exists()
    with caplog.at_level(logging.INFO):
        oxy_params = oxy_fitting.gather_oxy_params(f_path)
        assert "Failed to load" in caplog.messages[0]
        assert oxy_params.isnull().values.all()

def test_calculate_weights():
    #   Set some quick db values
    pressure = np.array([50, 150, 250, 400, 800, 1500, 2500, 5000])
    p_bins = [
        0,
        100,
        100 + 1e-5, #   Epsilon = 1e-5
        300,
        300 + 1e-5,
        500,
        500 + 1e-5,
        1200,
        1200 + 1e-5,
        2000,
        2000 + 1e-5,
        7000,
    ]
    w_bins = [20, 20, 25, 25, 50, 50, 100, 100, 200, 200, 500, 500]
    wgt_manual = scipy.interpolate.interp1d(p_bins, w_bins)(pressure)   #   Pull weights out

    wgt = oxy_fitting.calculate_weights(pressure)

    assert np.array_equal(wgt, wgt_manual)


"""
Classes, definitions and utilities for use across ctdcal fitting modules.
"""
import json

from munch import Munch


class BottleFlags(Munch):
    """
    A dictionary class with the attribute-style access of Munch, plus methods
    for adding nodes and flag data, and loading or saving to/from a JSON file.
    TODO: Move this to ctdcal.flagging.common after reorg
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_node(self, label, keys):
        """
        Add a new empty node.

        Parameters
        ----------
        label : str
            name of node
        keys : list
            names of node keys
        """
        key_dict = {k: [] for k in keys}
        node = ({label: BottleFlags(key_dict)})
        self.update(node)

    def update_node(self, **kwargs):
        """
        Add a row of values to a node. Requires an existing node with keys
        that are already defined.

        Parameters
        ----------
        kwargs
            Named keyword arguments as required for a node.

        """
        for k, v in kwargs.items():
            self[k].append(v)

    def save(self, fname):
        """
        Export the flags to a JSON file. Existing contents of the file
        will be overwritten. The file must already exist.

        Pretty printing may be removed or made optional in future updates.

        Parameters
        ----------
        fname : str or Path-like
            filename
        """
        with open(fname, 'w') as f:
            f.write(self.toJSON(indent=4))


class NodeNotFoundError(Exception):
    pass


# Function definitions
# --------------------

# BottleFlag wrangling
# TODO: Move to ctdcal.flagging.common after reorg
def df_node_to_BottleFlags(df):
    """
    Convert a flag node from a DataFrame to a formatted BottleFlags object

    Parameters
    ----------
    df : DataFrame
        Node data to convert.

    Returns
    -------
    BottleFlags object
    """
    node_dict = df.to_dict()
    for k, v in node_dict.items():
        node_dict[k] = [vv for kk, vv in v.items()]
    return BottleFlags(node_dict)


def get_node(fname, label):
    """
    Return a node from a BottleFlags file.

    Parameters
    ----------
    fname : str or Path-like
        Name of the BottleFlags file.
    label : str
        Name of the node to return.

    Returns
    -------
    BottleFlags object
    """
    with open(fname, 'r') as f:
        flags = json.load(f)
        if label in flags:
            return BottleFlags(flags[label])
        else:
            raise NodeNotFoundError


def save_node(fname, node, label, create_new=False):
    """
    Save an updated node to a BottleFlags file. The file must already exist.
    Optionally create a new node if the named node does not exist.

    Parameters
    ----------
    fname : str or Path-like
        Name of the BottleFlags file.
    node : BottleFlags object
        The formatted node data.
    label : str
        Name of the node to save.
    create_new : Bool
        If true, create a new node if it does not exist. Default is false.

    """
    with open(fname, 'r') as f:
        buf = f.read()
    if buf == '':
        # File exists but is empty
        flags = BottleFlags()
    else:
        flags = BottleFlags.fromJSON(buf)
    if label in flags or create_new is True:
        flags[label] = node
    else:
        raise NodeNotFoundError("The node '%s' was not found in %s" % (label, fname))
    flags.save(fname)



"""
Process all CTD and bottle data using ODF routines.
"""

# import needed ctdcal modules
from ctdcal import (
    convert,
    fit_ctd,
    odf_io,
    oxy_fitting,
    process_bottle,
    process_ctd,
    rinko,
)
from ctdcal.common import load_user_config, validate_file

import logging


log = logging.getLogger(__name__)

USERCONFIG = 'ctdcal/cfg.yaml'
user_cfg = load_user_config(validate_file(USERCONFIG))


def odf_process_all():

    #####
    # Step 0: Load and define necessary variables
    #####

    # cfg = get_ctdcal_config()

    #####
    # Step 1: Generate intermediate file formats (.pkl, _salts.csv, _reft.csv)
    #####

    # load station/cast list from file
    try:
        ssscc_list = process_ctd.get_ssscc_list()
    except FileNotFoundError:
        log.info("No ssscc.csv file found, generating from .hex file list")
        ssscc_list = process_ctd.make_ssscc_list()

    # convert raw .hex files
    convert.hex_to_ctd(ssscc_list)

    # process time files
    convert.make_time_files(ssscc_list, user_cfg.datadir, user_cfg)

    # process bottle file
    convert.make_btl_mean(ssscc_list)

    # generate salt .csv files
    odf_io.process_salts(ssscc_list, user_cfg)

    # generate reftemp .csv files
    process_bottle.process_reft(ssscc_list)

    #####
    # Step 2: calibrate pressure, temperature, conductivity, and oxygen
    #####

    # load in all bottle and time data into DataFrame
    time_data_all = process_ctd.load_all_ctd_files(ssscc_list)
    btl_data_all = process_bottle.load_all_btl_files(ssscc_list)

    # process pressure offset
    # TODO: these functions return an updated dataframe, which we aren't
    #   assigning or reassigning to anything. Instead we trust that the
    #   updates which happen in the other module are visible by this one
    #   too (they  indeed seem to be). Is this a safe assumption?
    process_ctd.apply_pressure_offset(btl_data_all)
    process_ctd.apply_pressure_offset(time_data_all)

    # create cast depth log file
    process_ctd.make_depth_log(time_data_all)

    # calibrate temperature against reference
    fit_ctd.calibrate_temp(btl_data_all, time_data_all)

    # calibrate conductivity against reference
    btl_data_all, time_data_all = fit_ctd.calibrate_cond(btl_data_all, time_data_all, user_cfg, 'salt')

    # calculate params needs for oxy/rinko calibration
    oxy_fitting.prepare_oxy(btl_data_all, time_data_all, ssscc_list, user_cfg, 'oxygen')

    # calibrate oxygen against reference
    oxy_fitting.calibrate_oxy(btl_data_all, time_data_all, ssscc_list)
    rinko.calibrate_oxy(btl_data_all, time_data_all, ssscc_list)

    #####
    # Step 3: export data
    #####

    # export files for making cruise report figs
    # process_bottle.export_report_data(btl_data_all)

    # export to Exchange format
    process_ctd.export_ct1(time_data_all, ssscc_list)
    process_bottle.export_hy1(btl_data_all)

    # run: ctd_to_bottle.py


if __name__ == "__main__":
    odf_process_all()



import sys
from pathlib import Path

import numpy as np
import pandas as pd
from bokeh.io import curdoc
from bokeh.layouts import column, row
from bokeh.models import (
    BoxSelectTool,
    Button,
    ColumnDataSource,
    DataTable,
    Div,
    MultiSelect,
    Range1d,
    Select,
    StringFormatter,
    TableColumn,
    TextInput,
)
from bokeh.plotting import figure

from ctdcal import get_ctdcal_config, io
from ctdcal.common import load_user_config, validate_file
from ctdcal.fitting.common import df_node_to_BottleFlags, get_node, save_node

cfg = get_ctdcal_config()
USERCONFIG = "ctdcal/cfg.yaml"
user_cfg = load_user_config(validate_file(USERCONFIG))
FLAGFILE = Path(user_cfg.datadir, "flag", user_cfg.bottleflags_man)

# TODO: abstract parts of this to a separate file
# TODO: following above, make parts reusable?

# load continuous CTD data and make into a dict (only ~20MB)
file_list = sorted(Path(cfg.dirs["pressure"]).glob("*ct1.csv"))
ssscc_list = [ssscc.stem[:5] for ssscc in file_list]
ctd_data = []
for f in file_list:
    print(f"Loading {f}")
    header, df = io.load_exchange_ctd(f)
    df["SSSCC"] = header["STNNBR"].zfill(3) + header["CASTNO"].zfill(2)
    ctd_data.append(df)
ctd_data = pd.concat(ctd_data, axis=0, sort=False)

# load bottle file
fname = list(Path(cfg.dirs["pressure"]).glob("*hy1.csv"))[0]
btl_data = io.load_exchange_btl(fname).replace(-999, np.nan)
btl_data["SSSCC"] = btl_data["STNNBR"].apply(lambda x: f"{x:03d}") + btl_data[
    "CASTNO"
].apply(lambda x: f"{x:02d}")

#   Create salinity residuals
btl_data["Residual"] = btl_data["SALNTY"] - btl_data["CTDSAL"]
btl_data[["CTDPRS", "Residual"]] = btl_data[["CTDPRS", "Residual"]].round(4)
btl_data["Comments"] = ""
btl_data["New Flag"] = btl_data["SALNTY_FLAG_W"].copy()

#   Temperature
btl_data["t_res"] = (btl_data["REFTMP"] - btl_data["CTDTMP"]).round(4)
btl_data["New T Flag"] = btl_data["REFTMP_FLAG_W"].copy()

#   Aaaand oxygen
if "OXYGEN" not in btl_data.columns:
    btl_data["OXYGEN"] = np.nan
    btl_data["OXYGEN_FLAG_W"] = 9
btl_data["o_res"] = (btl_data["OXYGEN"] - btl_data["CTDOXY"]).round(4)
btl_data["New O Flag"] = btl_data["OXYGEN_FLAG_W"].copy()

deltas_d = {"CTDSAL": "Residual", "CTDTMP": "t_res", "CTDOXY": "o_res"}

# update with old handcoded flags if file exists
if FLAGFILE.exists():
    salt_flags_manual = get_node(FLAGFILE, "salt")
    salt_flags_manual_df = pd.DataFrame.from_dict(salt_flags_manual)
    salt_flags_manual_df = salt_flags_manual_df.rename(
        columns={
            "value": "New Flag",
            "cast_id": "SSSCC",
            "bottle_num": "SAMPNO",
            "notes": "Comments",
        }
    )

    # there's gotta be a better way... but this is good enough for now
    btl_data = btl_data.merge(salt_flags_manual_df, on=["SSSCC", "SAMPNO"], how="left")
    merge_rows = ~btl_data["New Flag_y"].isnull() | ~btl_data["Comments_y"].isnull()
    btl_data.loc[merge_rows, "New Flag_x"] = btl_data.loc[merge_rows, "New Flag_y"]
    btl_data.loc[merge_rows, "Comments_x"] = btl_data.loc[merge_rows, "Comments_y"]
    btl_data = btl_data.rename(
        columns={"New Flag_x": "New Flag", "Comments_x": "Comments"}
    ).drop(columns=["New Flag_y", "Comments_y"])

# make downcast data point by interpolating bottle points on CTD data
downcast_data = []
for ssscc in ssscc_list:
    btl_rows = btl_data["SSSCC"] == ssscc
    ctd_rows = ctd_data["SSSCC"] == ssscc
    df = pd.DataFrame()
    for v in ["CTDTMP", "CTDSAL", "CTDOXY"]:
        df["CTDPRS"] = btl_data.loc[btl_rows, "CTDPRS"]
        df[v] = np.interp(
            btl_data.loc[btl_rows, "CTDPRS"],
            ctd_data.loc[ctd_rows, "CTDPRS"],
            ctd_data.loc[ctd_rows, v],
        )
    downcast_data.append(df)

downcast_data = pd.concat(downcast_data)

# intialize widgets
parameter = Select(
    title="Parameter", options=["CTDSAL", "CTDTMP", "CTDOXY"], value="CTDSAL"
)
ref_dict = {"CTDSAL": "SALNTY", "CTDTMP": "REFTMP", "CTDOXY": "OXYGEN"}
ref_param = Select(
    title="Reference",
    options=["SALNTY", "REFTMP", "OXYGEN"],
    value=ref_dict[parameter.value],
    disabled=True,
)

station = Select(title="Station", options=ssscc_list, value=ssscc_list[0])

# explanation of flags:
# https://cchdo.github.io/hdo-assets/documentation/manuals/pdf/90_1/chap4.pdf
flag_list = MultiSelect(
    title="Plot data flagged as:",
    value=["1", "2", "3", "9"],
    options=[
        ("1", "1 [Uncalibrated]"),
        ("2", "2 [Acceptable]"),
        ("3", "3 [Questionable]"),
        ("4", "4 [Bad]"),
        ("9", "9 [No Data]"),
    ],
    min_height=98,
)
# returns list of select options, e.g., ['2'] or ['1','2']
flag_input = Select(
    title="Flag:",
    options=[
        ("1", "1 [Uncalibrated]"),
        ("2", "2 [Acceptable]"),
        ("3", "3 [Questionable]"),
        ("4", "4 [Bad]"),
    ],
    value="3",
)
comment_box = TextInput(value="", title="Comment:")

# button_type: default, primary, success, warning or danger
flag_button = Button(label="Apply to selected", button_type="primary")
comment_button = Button(label="Apply to selected", button_type="warning")
save_button = Button(label="Save flagged data", button_type="success")
exit_button = Button(label="Exit flagging tool", button_type="danger")

vspace = Div(text=""" """, width=200, height=50)
residual_guide_text = Div(
    text="""<br><br>
    <u>Questionable Limits</u><br>
    <b>0.020</b> [0-500 m]<br>
    <b>0.010</b> [500-1000 m]<br>
    <b>0.005</b> [1000-2000 m]<br>
    <b>0.002</b> [2000-6000 m]""",
    width=150,
    height=135,
)
bulk_flag_text = Div(
    text="""<br><br>
    <b>Bulk Bottle Flagging:</b><br>
    Select multiple bottles using the table
    (with shift/control) or the 'Box Select' tool on the plot.""",
    width=150,
    height=135,
)

# set up datasources
src_table = ColumnDataSource(data=dict())
src_table_changes = ColumnDataSource(data=dict())
src_plot_trace = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_downcast = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_upcast = ColumnDataSource(data=dict(x=[], y=[]))
src_plot_btl = ColumnDataSource(data=dict(x=[], y=[]))

# set up plots
fig = figure(
    height=900,
    width=400,
    title="{} vs CTDPRS [Station {}]".format(parameter.value, station.value),
    tools="pan,box_zoom,wheel_zoom,box_select,reset",
    y_axis_label="Pressure (dbar)",
)
fig.line(
    "x",
    "y",
    line_color="#000000",
    line_width=2,
    source=src_plot_trace,
    legend_label="CTD Trace",
)
btl_sal = fig.asterisk(
    "x",
    "y",
    size=12,
    line_width=1.5,
    color="#0033CC",
    source=src_plot_btl,
    legend_label="Bottle sample",
)
ctd_sal = fig.circle(
    "x",
    "y",
    size=7,
    color="#BB0000",
    source=src_plot_downcast,
    legend_label="Downcast CTD sample",
)
upcast_sal = fig.triangle(
    "x",
    "y",
    size=7,
    color="#00BB00",
    source=src_plot_upcast,
    legend_label="Upcast CTD sample",
)
fig.select(BoxSelectTool).continuous = False
# fig.y_range = Range1d(max_y_value, 0)
fig.y_range.flipped = True  # invert y-axis
fig.legend.location = "bottom_right"
fig.legend.border_line_width = 3
fig.legend.border_line_alpha = 1
btl_sal.nonselection_glyph.line_alpha = 0.2
ctd_sal.nonselection_glyph.fill_alpha = 1  # makes CTDSAL *not* change on select
upcast_sal.nonselection_glyph.fill_alpha = 1  # makes CTDSAL *not* change on select

threshes = {
    "CTDSAL": np.array([0.002, 0.005, 0.010, 0.020]),
    "CTDTMP": np.array([0.002, 0.005, 0.010, 0.020]),
    "CTDOXY": np.array([0.625, 1.250, 2.500, 5.000]),   #   Not advised to follow this - biology happens
}

#   Residuals plot
src_plot_btl_del = ColumnDataSource(data=dict(x=[], y=[]))
fig2 = figure(
    height=900,
    width=400,
    title="{} residual vs CTDPRS [Station {}]".format(parameter.value, station.value),
    tools="pan,box_zoom,wheel_zoom,box_select,reset",
    # y_axis_label="Pressure (dbar)",
    y_range=fig.y_range,
)
# thresh = np.array([0.002, 0.005, 0.010, 0.020])
thresh = threshes[parameter.value]
p_range = np.array([6000, 2000, 1000, 500])
thresh = np.append(thresh, thresh[-1])
p_range = np.append(p_range, 0)
btl_sal2 = fig2.asterisk(
    "x",
    "y",
    size=12,
    line_width=1.5,
    color="#0033CC",
    source=src_plot_btl_del,
)
fig2.step(thresh, p_range)
fig2.step(-thresh, p_range)
fig2.select(BoxSelectTool).continuous = False
fig2.y_range = fig.y_range
# fig2.y_range.flipped = True  # invert y-axis

# define callback functions


def update_selectors():

    print("exec update_selectors()")
    ctd_rows = ctd_data["SSSCC"] == station.value
    table_rows = btl_data["SSSCC"] == station.value
    btl_rows = (btl_data["New Flag"].isin([int(v) for v in flag_list.value])) & (
        btl_data["SSSCC"] == station.value
    )

    # update table data
    current_table = btl_data[table_rows].reset_index()

    # print("Parameter: ", parameter.value, current_table[parameter.value][0])
    ref_param.value = ref_dict[parameter.value]
    # print("Reference parameter: ", ref_param.value, current_table[ref_param.value][0])

    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "t_res": current_table["t_res"],
        "o_res": current_table["o_res"],
        "CTD Param": current_table[parameter.value].round(4),
        "Reference": current_table[ref_param.value].round(4),
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
    }

    # update plot data
    src_plot_trace.data = {
        "x": ctd_data.loc[ctd_rows, parameter.value],
        "y": ctd_data.loc[ctd_rows, "CTDPRS"],
    }
    src_plot_upcast.data = {
        "x": btl_data.loc[table_rows, parameter.value],
        "y": btl_data.loc[table_rows, "CTDPRS"],
    }
    src_plot_downcast.data = {
        "x": downcast_data.loc[table_rows, parameter.value],
        "y": downcast_data.loc[table_rows, "CTDPRS"],
    }
    src_plot_btl.data = {
        "x": btl_data.loc[btl_rows, ref_param.value],
        "y": btl_data.loc[btl_rows, "CTDPRS"],
    }
    src_plot_btl_del.data = {
        "x": btl_data.loc[btl_rows, deltas_d[parameter.value]],
        "y": btl_data.loc[btl_rows, "CTDPRS"],
    }

    # update plot labels/axlims
    fig.title.text = "{} vs CTDPRS [Station {}]".format(parameter.value, station.value)
    fig.xaxis.axis_label = parameter.value

    fig2.title.text = "{} Residual".format(parameter.value)
    fig2.xaxis.axis_label = parameter.value

    # Set the y-range from 0 to the maximum value
    max_y = ctd_data.loc[ctd_rows, "CTDPRS"].max()
    fig.y_range = Range1d(max_y + 0.05 * max_y, 0)
    fig2.y_range = fig.y_range

    # Reset the x-axes when plots refresh
    max_x_fig = ctd_data.loc[ctd_rows, parameter.value].max()
    min_x_fig = ctd_data.loc[ctd_rows, parameter.value].min()
    fig.x_range = Range1d(min_x_fig, max_x_fig)

    max_x_fig2 = max(btl_data.loc[btl_rows, deltas_d[parameter.value]].max(), 0.025)
    min_x_fig2 = min(btl_data.loc[btl_rows, deltas_d[parameter.value]].min(), -0.025)
    fig2.x_range = Range1d(min_x_fig2, max_x_fig2)

    # deselect all datapoints
    btl_sal.data_source.selected.indices = []
    src_table.selected.indices = []


def edit_flag():

    print("exec edit_flag()")

    btl_data.loc[
        btl_data["SSSCC"] == src_table.data["SSSCC"].values[0],
        "New Flag",
    ] = src_table.data["flag"].values
    btl_data.loc[
        btl_data["SSSCC"] == src_table.data["SSSCC"].values[0],
        "Comments",
    ] = src_table.data["Comments"].values

    edited_rows = (btl_data["New Flag"].isin([3, 4])) | (btl_data["Comments"] != "")

    src_table_changes.data = {
        "SSSCC": btl_data.loc[edited_rows, "SSSCC"],
        "SAMPNO": btl_data.loc[edited_rows, "SAMPNO"],
        "diff": btl_data.loc[edited_rows, "Residual"],
        "flag_old": btl_data.loc[edited_rows, "SALNTY_FLAG_W"],
        "flag_new": btl_data.loc[edited_rows, "New Flag"],
        "Comments": btl_data.loc[edited_rows, "Comments"],
    }


def apply_flag():

    print("Applying flags")

    table_rows = btl_data["SSSCC"] == station.value
    selected_rows = src_table.selected.indices

    # update table data
    current_table = btl_data[table_rows].reset_index()
    current_table.loc[selected_rows, "New Flag"] = int(flag_input.value)
    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "CTDSAL": current_table["CTDSAL"],
        "SALNTY": current_table["SALNTY"],
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
        "CTD Param": current_table[parameter.value].round(4),
        "Reference": current_table[ref_param.value].round(4),
        "o_res": current_table["o_res"],
        "t_res": current_table["t_res"],
    }


def apply_comment():

    print("Applying Comments")

    table_rows = btl_data["SSSCC"] == station.value
    selected_rows = src_table.selected.indices

    # update table data
    current_table = btl_data[table_rows].reset_index()
    current_table.loc[selected_rows, "Comments"] = comment_box.value
    src_table.data = {  # this causes edit_flag() to execute
        "SSSCC": current_table["SSSCC"],
        "SAMPNO": current_table["SAMPNO"],
        "CTDPRS": current_table["CTDPRS"],
        "CTDSAL": current_table["CTDSAL"],
        "SALNTY": current_table["SALNTY"],
        "diff": current_table["Residual"],
        "flag": current_table["New Flag"],
        "Comments": current_table["Comments"],
        "CTD Param": current_table[parameter.value].round(4),
        "Reference": current_table[ref_param.value].round(4),
        "o_res": current_table["o_res"],
        "t_res": current_table["t_res"],
    }


def save_data():

    print("Saving flagged data...")

    # get data from table
    df_out = pd.DataFrame.from_dict(src_table_changes.data)

    # minor changes to columns/names/etc.
    df_out = df_out.rename(
        columns={
            "flag_new": "value",
            "SSSCC": "cast_id",
            "SAMPNO": "bottle_num",
            "Comments": "notes",
        }
    ).drop(columns=["flag_old", "diff"])

    # save it
    salt = df_node_to_BottleFlags(df_out)
    flagfile = validate_file(FLAGFILE, create=True)
    save_node(flagfile, salt, "salt", create_new=True)


def exit_bokeh():
    print("Stopping flagging software")
    sys.exit()


def selected_from_plot(attr, old, new):

    # update using bottle number, not index
    # currently there is a bug if not all data from a cast are plotted
    src_table.selected.indices = new


def selected_from_table(attr, old, new):

    # update using bottle number, not index
    # currently there is a bug if not all data from a cast are plotted
    btl_sal.data_source.selected.indices = new


# set up change callbacks
parameter.on_change("value", lambda attr, old, new: update_selectors())
station.on_change("value", lambda attr, old, new: update_selectors())
flag_list.on_change("value", lambda attr, old, new: update_selectors())
flag_button.on_click(apply_flag)
comment_button.on_click(apply_comment)
save_button.on_click(save_data)
exit_button.on_click(exit_bokeh)
src_table.on_change("data", lambda attr, old, new: edit_flag())
src_table.selected.on_change("indices", selected_from_table)
btl_sal.data_source.selected.on_change("indices", selected_from_plot)


# build data tables
columns = []
fields = [
    "SSSCC",
    "SAMPNO",
    "CTDPRS",
    "CTD Param",
    "Reference",
    "t_res",
    "diff",
    "o_res",
    "flag",
    "Comments",
]
ref_dict
titles = [
    "SSSCC",
    "Bottle",
    "CTDPRS",
    "CTD Param",
    "Reference",
    "t_res",
    "s_res",
    "o_res",
    "Flag",
    "Comments",
]
widths = [50, 40, 65, 65, 65, 65, 65, 65, 15, 200]
for field, title, width in zip(fields, titles, widths):
    if field == "flag":
        strfmt_in = {"text_align": "center", "font_style": "bold"}
    elif field == "Comments":
        strfmt_in = {}
    else:
        strfmt_in = {"text_align": "right"}
    columns.append(
        TableColumn(
            field=field,
            title=title,
            width=width,
            formatter=StringFormatter(**strfmt_in),
        )
    )

columns_changed = []
fields = ["SSSCC", "SAMPNO", "diff", "flag_old", "flag_new", "Comments"]
titles = ["SSSCC", "Bottle", "Residual", "Old", "New", "Comments"]
widths = [50, 40, 65, 15, 15, 375]
for field, title, width in zip(fields, titles, widths):
    if field == "flag_old":
        strfmt_in = {"text_align": "center", "font_style": "bold"}
    elif field == "flag_new":
        strfmt_in = {"text_align": "center", "font_style": "bold", "text_color": "red"}
    elif field == "Comments":
        strfmt_in = {}
    else:
        strfmt_in = {"text_align": "right"}
    columns_changed.append(
        TableColumn(
            field=field,
            title=title,
            width=width,
            formatter=StringFormatter(**strfmt_in),
        )
    )

data_table = DataTable(
    source=src_table,
    columns=columns,
    index_width=20,
    #   width is originally 565 + 50 + 20 = 635
    width=565 + 150 + 20,  # sum of col widths + fudge factor + idx width
    height=600,
    editable=True,
    fit_columns=False,
    sortable=False,
)
data_table_changed = DataTable(
    source=src_table_changes,
    columns=columns_changed,
    index_width=20,
    width=565 + 50 + 20,  # sum of col widths + fudge factor + idx width
    height=200,
    editable=False,
    fit_columns=False,
    sortable=False,
)
data_table_title = Div(text="""<b>All Station Data:</b>""", width=200, height=15)
data_table_changed_title = Div(
    text="""<b>Flagged Data:</b>""", width=200, height=15
)

controls = column(
    parameter,
    ref_param,
    station,
    flag_list,
    residual_guide_text,
    bulk_flag_text,
    flag_input,
    flag_button,
    comment_box,
    comment_button,
    vspace,
    save_button,
    exit_button,
    width=170,
)
tables = column(
    data_table_title, data_table, data_table_changed_title, data_table_changed
)

curdoc().add_root(row(controls, tables, fig, fig2))
curdoc().title = "CTDO Data Flagging Tool"

update_selectors()



"""
Classes, definitions and utilities for all ctdcal modules
"""
from pathlib import Path
import yaml

from munch import munchify


# Function definitions
# --------------------

# Configuration
def load_user_config(cfgfile):
    """
    Load user-defined parameters from a configuration file. Return a Munch
    object (dictionary).

    Parameters
    ----------
    cfgfile : str or Path-like
        Path to the configuration file.

    Returns
    -------
    Munch object
    """
    with open(cfgfile, 'r') as f:
        cfg = yaml.safe_load(f)
        return munchify(cfg)


# Input Validation
def validate_dir(pathname, create=False):
    """
    Test if a directory exists, and optionally create it if it does not. Raise
    an exception if the directory does not exist, cannot be created, or is not
    a directory. Return the validated path.

    Parameters
    ----------
    pathname : str or PathLike
        Directory to validate.
    create : bool
        If true, create the directory if it does not exist. Default is false.

    Returns
    -------
    Path object
    """
    p = Path(pathname)
    if create is True:
        p.mkdir(parents=True, exist_ok=True)
        return p
    elif p.is_dir():
        return p
    elif p.exists():
        raise FileExistsError("%s already exists but is not a directory." % str(p))
    else:
        raise FileNotFoundError("The directory %s could not be found" % str(p))


def validate_file(pathname, create=False):
    """
    Test if a file exists, and optionally create it if it does not. Raise
    an exception if the file does not exist, cannot be created, or is not
    a file. Return the validated path.

    Parameters
    ----------
    pathname : str or Path-like
        Filename to validate.
    create : bool
        If true, create the file if it does not exist. Default is false.

    Returns
    -------
    Path object
    """
    p = Path(pathname)
    if create is True:
        p.parent.mkdir(parents=True, exist_ok=True)
        if p.is_dir():
            raise FileExistsError
        p.touch(exist_ok=True)
        return p
    elif p.is_file():
        return p
    elif p.exists():
        raise FileExistsError("%s exists but is not a file." % str(p))
    else:
        raise FileNotFoundError("The file %s could not be found." % str(p))



#! /usr/bin/env python
import argparse
import os
import sys

import pandas as pd

from . import process_bottle as btl

# File extension to use for output files (csv-formatted)
FILE_EXT = "csv"
EXT_PKL = "pkl"

# File extension to use for converted output
CONVERTED_SUFFIX = "_cnv"

# File extension to use for processed output
BOTTLE_SUFFIX = "_btl"
MEAN_SUFFIX = "_mean"
MEDIAN_SUFFIX = "_median"

# Column name for the bottle fire number
BOTTLE_FIRE_NUM_COL = "btl_fire_num"

DEBUG = False


def debugPrint(*args, **kwargs):
    if DEBUG:
        errPrint(*args, **kwargs)


def errPrint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


# -------------------------------------------------------------------------------------
# Main function of the script should it be run as a stand-alone utility.
# -------------------------------------------------------------------------------------
def main(argv):
    # MK: depreciated 04/27/20, use ctdcal.convert.make_btl_mean instead

    parser = argparse.ArgumentParser(
        description="Process bottle fire data from a converted, csv-formatted text file"
    )
    parser.add_argument(
        "cnv_file", metavar="cnv_file", help="the converted data file to process"
    )

    # debug messages
    parser.add_argument(
        "-d", "--debug", action="store_true", help="display debug messages"
    )

    # output directory
    parser.add_argument(
        "-o", metavar="dest_dir", dest="outDir", help="location to save output files"
    )

    args = parser.parse_args()
    if args.debug:
        global DEBUG
        DEBUG = True
        debugPrint("Running in debug mode")

    # Verify converted exists
    if not os.path.isfile(args.cnv_file):
        errPrint("ERROR: Input converted file:", args.cnv_file, "not found\n")
        sys.exit(1)

    # Set the default output directory to be the same directory as the hex file
    outputDir = os.path.dirname(args.cnv_file)

    # Used later for building output file names
    filename_ext = os.path.basename(args.cnv_file)  # original filename with ext
    filename_base = os.path.splitext(filename_ext)[0]  # original filename w/o ext

    # Verify Output Directory exists
    if args.outDir:
        if os.path.isdir(args.outDir):
            outputDir = args.outDir
        else:
            errPrint("ERROR: Output directory:", args.outDir, "is unreachable.\n")
            sys.exit(1)

    debugPrint("Import converted data to dataframe... ", end="")
    imported_df = pd.read_pickle(args.cnv_file)
    debugPrint("Success!")

    debugPrint(imported_df.head())
    # debugPrint(imported_df.dtypes)

    # Retrieve the rows from the imported dataframe where the btl_fire_bool column == True
    # Returns a new dataframe
    bottle_df = btl.retrieveBottleData(imported_df)

    if bottle_df.empty:
        errPrint("Bottle fire data not found in converted file")
        sys.exit(1)
    else:
        total_bottles_fired = bottle_df[BOTTLE_FIRE_NUM_COL].iloc[-1]
        debugPrint(total_bottles_fired, "bottle fire(s) detected")
        bottle_num = 1

        while bottle_num <= total_bottles_fired:
            debugPrint("Bottle:", bottle_num)
            debugPrint(
                bottle_df.loc[bottle_df[BOTTLE_FIRE_NUM_COL] == bottle_num].head()
            )
            bottle_num += 1

    # # Build the filename for the bottle fire data
    # bottlefileName  = filename_base.replace(CONVERTED_SUFFIX,'') + BOTTLE_SUFFIX + '.' + FILE_EXT
    # bottlefilePath = os.path.join(outputDir, bottlefileName)
    #
    # # Save the bottle fire dataframe to file.
    # debugPrint('Saving bottle data to:', bottlefilePath + '... ', end='')
    # if not cnv.saveConvertedDataToFile(bottle_df, bottlefilePath, DEBUG):
    #     errPrint('ERROR: Could not save bottle fire data to file')
    # else:
    #     debugPrint('Success!')
    #
    mean_df = btl.bottle_mean(bottle_df)

    # Build the filename for the bottle fire mean data
    # meanfileName  = filename_base.replace(CONVERTED_SUFFIX,'') + BOTTLE_SUFFIX + MEAN_SUFFIX + '.' + FILE_EXT
    meanfileName = (
        filename_base.replace(CONVERTED_SUFFIX, "")
        + BOTTLE_SUFFIX
        + MEAN_SUFFIX
        + "."
        + EXT_PKL
    )
    meanfilePath = os.path.join(outputDir, meanfileName)

    # Save the bottle fire mean dataframe to file.
    debugPrint("Saving mean data to:", meanfilePath + "... ", end="")
    if not mean_df.to_pickle(meanfilePath):
        errPrint("ERROR: Could not save mean fire data to file")
    else:
        debugPrint("Success!")

    #### Statistical analysis later to see if mean or median is better for values

    # median_df = btl.bottle_median(bottle_df)
    #
    # # Build the filename for the bottle fire median data
    # medianfileName  = filename_base.replace(CONVERTED_SUFFIX,'') + BOTTLE_SUFFIX + MEDIAN_SUFFIX + '.' + FILE_EXT
    # medianfilePath = os.path.join(outputDir, medianfileName)
    #
    # # Save the bottle fire median dataframe to file.
    # debugPrint('Saving median data to:', medianfilePath + '... ', end='')
    # if not cnv.saveConvertedDataToFile(median_df, medianfilePath, DEBUG):
    #     errPrint('ERROR: Could not save median fire data to file')
    # else:
    #     debugPrint('Success!')


# -------------------------------------------------------------------------------------
# Required python code for running the script as a stand-alone utility
# -------------------------------------------------------------------------------------
if __name__ == "__main__":
    main(sys.argv[1:])



import logging
from zipfile import ZIP_DEFLATED, ZipFile, ZipInfo
from zipimport import ZipImportError

import numpy as np
import pytest
import requests

from ctdcal import io


def check_type(to_check, sub_dtype):
    """
    Type comparisons can fail on Windows due to int32/int64 behavior.
    This function checks subtypes so that tests don't fail on Windows.

    See https://stackoverflow.com/q/64901822/13155619 for more info
    """
    if sub_dtype is int:
        np_type = np.integer
    elif sub_dtype is float:
        np_type = np.floating

    return all([np.issubdtype(dtype, np_type) for dtype in to_check.dtypes])


def test_load_cnv(tmp_path):
    # make fake/empty .cnv file
    content = [
        "* header rows\n* header rows\n* header rows\n* header rows\n",
        "# nquan = 3\n# nvalues = 4\n# units = specified\n",
        "# name 0 = prDM\n# name 1 = depSM\n# name 2 = t090C\n",
        "# bad_flag = -9.990e-29\n",
        "# comment rows\n# comment rows\n# comment rows\n",
        "*END*\n",
        "      4.000      3.962     8.9121\n",
        "      6.000      5.943     8.9125\n",
        "      8.000      7.924     8.9123\n",
        "     10.000      9.905     8.9126\n",
    ]

    # write fake data and check read from file
    with open(tmp_path / "test_1.cnv", "w+") as f:
        for line in content:
            f.write(line)

    cnv = io.load_cnv(tmp_path / "test_1.cnv")
    assert cnv.shape == (4, 3)
    assert check_type(cnv, float)
    assert cnv.columns.tolist() == ["prDM", "depSM", "t090C"]


def test_load_exchange_btl(caplog, tmp_path, monkeypatch):
    # make fake/empty Exchange file
    content = [
        "BOTTLE,20210101ODFSIO\n",
        "# comment about EXPOCODE\n",
        "#\n#\n#\n#\n",
        "EXPOCODE,SECT_ID,STNNBR,CASTNO,SAMPNO\n",
        ",,,,\n",
        "  123420210101,   ABC,     1,  2,      3\n",
        "  123420210101,   ABC,     1,  2,      4\n",
        "END_DATA",
    ]

    # write fake data and check read from file
    fname = "test_hy1.csv"
    with open(tmp_path / fname, "w+") as f:
        for line in content:
            f.write(line)

    with caplog.at_level(logging.INFO):
        hy1 = io.load_exchange_btl(tmp_path / fname)
    assert hy1.shape == (2, 5)
    assert check_type(hy1[["EXPOCODE", "STNNBR", "CASTNO", "SAMPNO"]], int)
    assert all(hy1["SECT_ID"] == "ABC")  # should not have any leading spaces
    assert f"{fname} from local file" in caplog.messages[0]

    # check read works with str path as well
    with caplog.at_level(logging.INFO):
        assert hy1.equals(io.load_exchange_btl(f"{str(tmp_path)}/{fname}"))
        assert f"{fname} from local file" in caplog.messages[1]

    # mock downloading data from CCHDO
    class MockResponse(object):
        def __init__(self):
            self.text = "".join(content)

    def mock_get(*args, **kwargs):
        return MockResponse()

    # "overwrite" requests.get() call in ctdcal.io with mock_get()
    monkeypatch.setattr(requests, "get", mock_get)

    # check read works from URL
    with caplog.at_level(logging.INFO):
        assert hy1.equals(io.load_exchange_btl(f"http://fakeurl/{fname}"))
        assert f"{fname} from http link" in caplog.messages[2]


def test_load_exchange_ctd(caplog, tmp_path, monkeypatch):
    # make fake/empty Exchange file
    content = [
        "CTD,20210101ODFSIO\n",
        "# comment about NUMBER_HEADERS\n",
        "# comment about CTDPRS\n",
        "#\n#\n#\n#\n",
        "NUMBER_HEADERS = 11\n",
        "EXPOCODE = 012345678910\n",
        "SECT_ID = ABC\n",
        "STNNBR = 1\n",
        "CASTNO = 1\n",
        "DATE = 20220101\n",
        "TIME = 1234\n",
        "LATITUDE = 45.6789\n",
        "LONGITUDE = -50.1234\n",
        "DEPTH = 96\n",
        "INSTRUMENT_ID = 987\n",
        "CTDPRS,CTDTMP,CTDTMP_FLAG_W,CTDSAL,CTDSAL_FLAG_W,CTDOXY,CTDOXY_FLAG_W\n",
        "DBAR,ITS-90,,PSS-78,,UMOL/KG,\n",
        "      0.0,   2.8000,1,  32.5320,2,    330.0,3\n",
        "      2.0,   2.8070,2,  33.0000,3,    200.0,4\n",
        "END_DATA",
    ]

    # write fake data and check read from file
    fname = "test_ct1.csv"
    with open(tmp_path / fname, "w+") as f:
        for line in content:
            f.write(line)

    with caplog.at_level(logging.INFO):
        header, ct1 = io.load_exchange_ctd(tmp_path / fname)
    assert ct1.shape == (2, 7)
    assert len(header) == 11
    assert check_type(ct1[["CTDPRS", "CTDTMP", "CTDSAL", "CTDOXY"]], float)
    assert check_type(ct1[["CTDTMP_FLAG_W", "CTDSAL_FLAG_W", "CTDOXY_FLAG_W"]], int)
    assert f"{fname} from local file" in caplog.messages[0]

    # check read works with str path as well
    caplog.clear()
    with caplog.at_level(logging.INFO):
        header_str, ct1_str = io.load_exchange_ctd(f"{str(tmp_path)}/{fname}")
        assert ct1.equals(ct1_str)
        assert header_str == header
        assert f"{fname} from local file" in caplog.messages[0]

    # mock downloading data from CCHDO (single cast file; .zip tested separately)
    class MockResponse(object):
        def __init__(self):
            self.content = "".join(content).encode("utf8")

    def mock_get(*args, **kwargs):
        return MockResponse()

    # "overwrite" requests.get() call in ctdcal.io with mock_get()
    monkeypatch.setattr(requests, "get", mock_get)

    # check read works from URL
    caplog.clear()
    with caplog.at_level(logging.INFO):
        header_url, ct1_url = io.load_exchange_ctd(f"http://fakeurl/{fname}")
        assert ct1.equals(ct1_url)
        assert header_url == header
        assert f"{fname} from http link" in caplog.messages[0]

    # write fake data (x3 files) to .zip
    zname = "test_ctd.zip"
    with ZipFile(tmp_path / zname, "w") as zf:
        for n in [1, 2, 3]:
            zf.writestr(
                ZipInfo(filename=f"CTD_{n}_ct1.csv"),
                "".join(content).encode("utf8"),
                compress_type=ZIP_DEFLATED,
            )

    # check read works from .zip
    caplog.clear()
    with caplog.at_level(logging.INFO):
        header_zip, ct1_zip = io.load_exchange_ctd(tmp_path / zname)
    assert "from local file" in caplog.messages[0]
    assert "from .zip" in caplog.messages[1]
    assert all(["open file object" in caplog.messages[n] for n in [2, 3, 4]])
    assert len(ct1_zip) == 3
    assert all([ct1_zip[n].equals(ct1) for n in [0, 1, 2]])
    assert all([header_zip[n] == header for n in [0, 1, 2]])

    # check reading specific # of files from .zip works
    caplog.clear()
    with caplog.at_level(logging.INFO):
        header_zip, zipped = io.load_exchange_ctd(tmp_path / zname, n_files=2)
    assert "from local file" in caplog.messages[0]
    assert "from .zip" in caplog.messages[1]
    assert all(["open file object" in caplog.messages[n] for n in [2, 3]])
    assert len(zipped) == 2
    assert all([zipped[n].equals(ct1) for n in [0, 1]])
    assert all([header_zip[n] == header for n in [0, 1]])

    # write nested zip (including "real" data is not important)
    with ZipFile(tmp_path / "level1.zip", "w") as zf1:
        zf1.writestr(ZipInfo(filename="test.csv"), " ".encode("utf8"))
    with ZipFile(tmp_path / "level0.zip", "w") as zf0:
        zf0.write(tmp_path / "level1.zip")

    # check error on recursive .zip
    with pytest.raises(ZipImportError, match="Recursive .zip files"):
        io.load_exchange_ctd(tmp_path / "level0.zip")



from ctdcal import flagging
import numpy as np
import pandas as pd
import pytest


# data = [np.nan] + 97 * [0] + [100, 100]
# data = pd.Series([np.nan] + 97 * [0] + [100, 100])
# data = pd.DataFrame([np.nan] + 97 * [0] + [100, 100])


def test_merge_flags():
    old_flags = 5 * [2]
    new_flags = 5 * [3]

    # check proper flags are kept
    np.testing.assert_array_equal(
        flagging._merge_flags(new_flags, old_flags, keep_higher=True), new_flags
    )
    np.testing.assert_array_equal(
        flagging._merge_flags(new_flags, old_flags, keep_higher=False), old_flags
    )

    # error if flag lengths are different
    with pytest.raises(ValueError):
        flagging._merge_flags(old_flags, new_flags[:-1])
    with pytest.raises(ValueError):
        flagging._merge_flags(old_flags[:-1], new_flags)


def test_nan_values():
    data = [0, 0, 0, np.nan]
    old_flags = len(data) * [2]

    # check NaNs are flagged properly (without old flags)
    assert all(flagging.nan_values(data)[:-1] == 2)
    assert flagging.nan_values(data)[-1] == 9

    # check re-defining flag_good/flag_bad (without old flags)
    new_flags = flagging.nan_values(data, flag_good=6, flag_nan=5)
    assert all(new_flags[:-1] == 6)
    assert new_flags[-1] == 5

    # check NaNs are flagged properly (with old flags)
    assert all(flagging.nan_values(data, old_flags=old_flags)[:-1] == 2)
    assert flagging.nan_values(data, old_flags=old_flags)[-1] == 9

    # check re-defining flag_good/flag_bad (with old flags)
    new_flags = flagging.nan_values(data, old_flags=old_flags, flag_good=6, flag_nan=5)
    assert all(new_flags[:-1] == 6)
    assert new_flags[-1] == 5


def test_outliers():
    data = [np.nan] + 97 * [0] + [100, 100]
    old_flags = (len(data) - 1) * [2] + [7]

    # check outliers are flagged properly (without old flags)
    assert all(flagging.outliers(data)[:-2] == 2)
    assert all(flagging.outliers(data)[-2:] == 4)

    # check re-defining flag_good/flag_bad (without old flags)
    new_flags = flagging.outliers(data, flag_good=6, flag_outlier=5)
    assert all(new_flags[:-2] == 6)
    assert all(new_flags[-2:] == 5)

    # check outliers are flagged properly (with old flags)
    assert all(flagging.outliers(data, old_flags=old_flags)[:-2] == 2)
    assert flagging.outliers(data, old_flags=old_flags)[-1] == old_flags[-1]
    assert flagging.outliers(data, old_flags=old_flags)[-2] == 4

    # check re-defining flag_good/flag_bad (with old_flags)
    new_flags = flagging.outliers(
        data, old_flags=old_flags, flag_good=6, flag_outlier=5
    )
    assert all(new_flags[:-2] == 6)
    assert new_flags[-1] == old_flags[-1]
    assert new_flags[-2] == 5

    # check outliers are not flagged with large n_sigma1
    assert all(flagging.outliers(data, n_sigma1=5)[:-2] == 2)
    assert all(flagging.outliers(data, n_sigma1=5)[-2:] == 4)
    assert all(flagging.outliers(data, n_sigma1=10) == 2)


def test_by_percent_diff():
    data = [np.nan] + 97 * [0] + [100, 100]
    ref = (len(data) - 1) * [0] + [80]
    old_flags = (len(data) - 1) * [2] + [7]

    # check large diffs are flagged properly (without old flags)
    assert all(flagging.by_percent_diff(data, ref)[:-2] == 2)
    assert all(flagging.by_percent_diff(data, ref)[-2:] == 3)

    # check re-defining flag_good/flag_bad (without old flags)
    new_flags = flagging.by_percent_diff(data, ref, flag_good=6, flag_bad=5)
    assert all(new_flags[:-2] == 6)
    assert all(new_flags[-2:] == 5)

    # check large diffs are flagged properly (with old flags)
    new_flags = flagging.by_percent_diff(data, ref, old_flags=old_flags)
    assert all(new_flags[:-2] == 2)
    assert new_flags[-1] == old_flags[-1]
    assert new_flags[-2] == 3

    # check re-defining flag_good/flag_bad (with old flags)
    new_flags = flagging.by_percent_diff(
        data, ref, old_flags=old_flags, flag_good=6, flag_bad=5
    )
    assert all(new_flags[:-2] == 6)
    assert new_flags[-1] == old_flags[-1]
    assert new_flags[-2] == 5

    # check large diffs are not flagged with sufficiently high thresh
    assert flagging.by_percent_diff(data, ref, percent_thresh=10)[-1] == 3
    assert flagging.by_percent_diff(data, ref, percent_thresh=20)[-1] == 2


@pytest.mark.parametrize(
    "offset, p", [(0.003, 2000), (0.006, 1000), (0.011, 500), (0.021, 0)]
)
def test_by_residual(offset, p):
    data = [np.nan] + 97 * [0] + [100, 100]
    ref = np.array(data) + offset
    pres = np.linspace(0, 3000, len(data))
    old_flags = (len(data) - 1) * [2] + [7]

    # check values above threshold are flagged properly (without old flags)
    new_flags = flagging.by_residual(data, ref, pres)
    assert all(new_flags[pres < p] == 2)
    assert all(new_flags[pres > p] == 3)

    # check re-defining flag_good/flag_bad (without old flags)
    new_flags = flagging.by_residual(data, ref, pres, flag_good=6, flag_bad=5)
    assert all(new_flags[pres < p] == 6)
    assert all(new_flags[pres > p] == 5)

    # check values above threshold are flagged properly (with old flags)
    new_flags = flagging.by_residual(data, ref, pres, old_flags=old_flags)
    assert all(new_flags[pres < p] == 2)
    assert all(new_flags[pres > p][:-1] == 3)
    assert new_flags[pres > p][-1] == old_flags[-1]

    # check re-defining flag_good/flag_bad (with old flags)
    new_flags = flagging.by_residual(
        data, ref, pres, old_flags=old_flags, flag_good=6, flag_bad=5
    )
    assert all(new_flags[pres < p] == 6)
    assert all(new_flags[pres > p][:-1] == 5)
    assert new_flags[pres > p][-1] == old_flags[-1]

    # error if threshold not one value longer than p_cutoff
    with pytest.raises(IndexError):
        flagging.by_residual([], [], [], threshold=[1, 2], p_cutoff=[100, 50, 0])
    with pytest.raises(IndexError):
        flagging.by_residual([], [], [], threshold=[1, 2, 3], p_cutoff=[100, 50, 0])
    with pytest.raises(IndexError):
        flagging.by_residual([], [], [], threshold=[1, 2, 3], p_cutoff=[100])

    # error if p_cutoff not in decreasing order
    with pytest.raises(ValueError):
        flagging.by_residual([], [], [], threshold=[1, 2, 3], p_cutoff=[0, 50])
    with pytest.raises(ValueError):
        flagging.by_residual([], [], [], threshold=[1, 2, 3, 4], p_cutoff=[100, 0, 50])
